from ._function import FunctionTool, function
from ._browser import BrowserTool
from ._agent import AgentTool, BrowserAgentTool

__all__ = [
    "function",
    "FunctionTool",
    "BrowserTool",
    "AgentTool",
    "BrowserAgentTool",
]



from dataclasses import dataclass
from typing import Optional, List, Dict, Any, Type
from pydantic import BaseModel

from npiai.types.shot import Shot


@dataclass(frozen=True)
class ToolMeta:
    name: str
    description: str
    schema: Dict[str, Any] = None
    model: Optional[Type[BaseModel]] = None
    few_shots: Optional[List[Shot]] = None



from typing import overload
from npiai import FunctionTool, BrowserTool, AgentTool, BrowserAgentTool, LLM


@overload
def wrap(tool: FunctionTool, llm: LLM = None) -> AgentTool: ...


@overload
def wrap(tool: BrowserTool, llm: LLM = None) -> BrowserAgentTool: ...


def wrap(
    tool: FunctionTool | BrowserTool, llm: LLM = None
) -> AgentTool | BrowserAgentTool:
    if isinstance(tool, BrowserTool):
        return BrowserAgentTool(tool, llm)

    if isinstance(tool, FunctionTool):
        return AgentTool(tool, llm)

    raise TypeError(f"app must be an instance of FunctionTool or BrowserTool")



import json
import os

from github import Github as PyGithub, Auth
from github.IssueComment import IssueComment
from github.Issue import Issue
from github.PullRequestComment import PullRequestComment
from github.PullRequest import PullRequest
from github.GithubObject import NotSet
from typing import Union, TypeVar, List, Literal

from npiai import FunctionTool, function, utils
from npiai.constant import app
from npiai.context import Context
from npiai.error.auth import UnauthorizedError

__PROMPT__ = """
You are a GitHub Agent helping users to manage their issues and pull requests.

## Example

Task: Get the latest issue from npi-ai/npi
Steps:
- get_issue_pr({ "query": "repo:npi-ai/npi is:issue", "max_results": 1 })

Task: Star and fork npi-ai/npi.
Steps:
- star({ "repo": "npi-ai/npi" })
- fork({ "repo": "npi-ai/npi" })

Task: Leave a comment in issue #42 from npi-ai/npi.
Steps:
- add_issue_comment({ "repo": "npi-ai/npi", id: 42, "body": "{{issue_comment}}" })
"""

_T = TypeVar("_T")


def _default_not_set(value: _T) -> _T | "NotSet":
    return value if value is not None else NotSet


class GitHub(FunctionTool):
    name = "github"
    description = "Manage GitHub issues and pull requests"
    system_prompt = __PROMPT__

    _token: str | None
    _client: PyGithub | None

    def __init__(self, access_token: str = None):
        super().__init__()
        self._token = access_token or os.environ.get("GITHUB_ACCESS_TOKEN", None)

    @classmethod
    def from_context(cls, ctx: Context) -> "GitHub":
        if not utils.is_cloud_env():
            raise RuntimeError(
                "GitHub tool can only be initialized from context in the NPi cloud environment"
            )
        creds = ctx.credentials(app_code=app.GITHUB)
        return GitHub(access_token=creds["access_token"])

    async def start(self):
        if self._token is None:
            raise UnauthorizedError("GitHub credentials are not found")
        self._client = PyGithub(auth=Auth.Token(self._token))
        await super().start()

    @staticmethod
    def _comment_to_json(comment: Union[IssueComment, PullRequestComment]):
        """
        Converts a Comment object to a JSON object
        Args:
            comment: an IssueComment or PullRequestComment object

        Returns:
            a JSON object
        """
        return {
            "id": comment.id,
            "created_at": comment.created_at.isoformat(),
            "updated_at": comment.updated_at.isoformat(),
            "author": comment.user.login,
            "body": comment.body,
        }

    @staticmethod
    def _issue_pr_to_json(issue_or_pr: Union[Issue, PullRequest]):
        """
        Converts an Issue object to a JSON object
        Args:
            issue_or_pr: an Issue or PullRequest object

        Returns:
            a JSON object
        """
        return {
            "number": issue_or_pr.number,
            "title": issue_or_pr.title,
            "created_at": issue_or_pr.created_at.isoformat(),
            "updated_at": issue_or_pr.updated_at.isoformat(),
            "is_closed": issue_or_pr.closed_at is not None,
            "author": issue_or_pr.user.login,
            "body": issue_or_pr.body,
            "comments_count": issue_or_pr.comments,
            "assignees": [user.login for user in issue_or_pr.assignees],
            "labels": [label.name for label in issue_or_pr.labels],
        }

    def _get_issue(self, repo_name: str, number: int) -> Issue:
        """
        Get an issue from the given repository

        Args:
            repo_name: name of the repository in format {owner}/{repo}
            number: issue number

        Returns:
            an Issue object
        """
        repo = self._client.get_repo(repo_name)
        return repo.get_issue(int(number))

    def _get_pull_request(self, repo_name: str, number: int) -> PullRequest:
        """
        Get a pull request from the given repository

        Args:
            repo_name: name of the repository in format {owner}/{repo}
            number: pull request number

        Returns:
            an Issue object
        """
        repo = self._client.get_repo(repo_name)
        return repo.get_pull(number)

    @function
    def star(self, repo: str):
        """
        Star a repository on GitHub.

        Args:
            repo: Name of the repository in format {owner}/{repo}
        """
        repo = self._client.get_repo(repo)
        user = self._client.get_user()
        user.add_to_starred(repo)

        return f"Starred {repo} on behalf of {user.login}"

    @function
    def fork(self, repo: str):
        """
        Fork a repository on GitHub.

        Args:
            repo: Name of the repository in format {owner}/{repo}
        """
        repo = self._client.get_repo(repo)
        user = self._client.get_user()
        forked = user.create_fork(repo)

        return f"Forked {repo} to {forked.full_name}"

    @function
    def watch(self, repo: str):
        """
        Subscribe to notifications (a.k.a. watch) for activity in a repository on GitHub.

        Args:
            repo: Name of the repository in format {owner}/{repo}
        """
        repo = self._client.get_repo(repo)
        user = self._client.get_user()
        user.add_to_watched(repo)

        return f"Watched {repo} on behalf of {user.login}"

    @function
    def search_repositories(self, query: str, max_results: int):
        """
        Search for repositories on GitHub.

        Args:
            query: Search query to search for GitHub repositories. Below are some query examples:
                1. Search for all repositories that contain "test" and use Python: `test language:python`
                2. Search for repositories with more than 1000 stars: `stars:>1000`
            max_results: Maximum number of results to return
        """
        res = self._client.search_repositories(query)

        results = []

        for repo in res[:max_results]:
            results.append(
                {
                    "name": repo.full_name,
                    "owner": repo.owner.login,
                    "url": repo.url,
                    "description": repo.description,
                    "topics": repo.topics,
                    "created_at": repo.created_at.isoformat(),
                    "updated_at": repo.updated_at.isoformat(),
                }
            )

        return json.dumps(results, ensure_ascii=False)

    @function
    def search_issue_pr(self, query: str, max_results: int):
        """
        Search for issues or pull requests on GitHub.

        Args:
            query: Search query to search for GitHub issues and pull requests. Below are some query examples:
                1. Search for all issues containing "performance" in the repository npi/npi: `is:issue repo:npi/npi performance`
                2. Search for open pull requests in repository npi/npi: `is:pr is:open repo:npi/npi`
            max_results: Maximum number of results to return
        """
        res = self._client.search_issues(query=query, sort="created")

        if res.totalCount == 0:
            return "No results found"

        results = []

        for item in res[:max_results]:
            results.append(self._issue_pr_to_json(item))

        return json.dumps(results, ensure_ascii=False)

    @function
    def get_issue(self, repo: str, number: int):
        """
        Get an issue from the given repository.

        Args:
            repo: Name of the repository in format {owner}/{repo}
            number: Issue number
        """
        issue = self._get_issue(repo_name=repo, number=number)
        return json.dumps(self._issue_pr_to_json(issue), ensure_ascii=False)

    @function
    def get_pull_request(self, repo: str, number: int):
        """
        Get a pull request from the given repository.

        Args:
            repo: Name of the repository in format {owner}/{repo}
            number: Pull request number
        """
        pr = self._get_pull_request(repo_name=repo, number=number)
        return json.dumps(self._issue_pr_to_json(pr), ensure_ascii=False)

    @function
    def create_issue(
        self,
        repo: str,
        title: str,
        body: str,
        labels: List[str] = None,
        assignees: List[str] = None,
    ):
        """
        Create an issue under the given repository.

        Args:
            repo: Name of the repository in format {owner}/{repo}
            title: Title of the issue
            body: Body of the issue in markdown format
            labels: List of labels to add to the issue
            assignees: List of users to assign to the issue
        """
        repository = self._client.get_repo(repo)
        issue = repository.create_issue(
            title=title,
            body=body,
            labels=_default_not_set(labels),
            assignees=_default_not_set(assignees),
        )

        return "Issue created:\n" + json.dumps(
            self._issue_pr_to_json(issue), ensure_ascii=False
        )

    @function
    def create_pull_request(
        self,
        repo: str,
        title: str,
        body: str,
        base: str,
        head: str,
        is_draft: bool = False,
        labels: List[str] = None,
        assignees: List[str] = None,
    ):
        """
        Create a pull request under the given repository.

        Args:
            repo: Name of the repository in format {owner}/{repo}
            title: Title of the pull request
            body: Body of the pull request
            base: Base branch of the pull request, i.e., the branch to merge the pull request into
            head: Head branch of the pull request, i.e., the branch with your changes
            is_draft: Whether the pull request is a draft or not
            labels: List of labels to add to the pull request
            assignees: List of users to assign to the pull request
        """
        repository = self._client.get_repo(repo)
        pr = repository.create_pull(
            base=base,
            title=title,
            body=body,
            head=head,
            draft=is_draft,
        )

        if labels:
            pr.add_to_labels(*labels)

        if assignees:
            pr.add_to_assignees(*assignees)

        return "Pull Request created:\n" + json.dumps(
            self._issue_pr_to_json(pr), ensure_ascii=False
        )

    @function
    def add_issue_comment(self, repo: str, number: int, body: str):
        """
        Add a comment to the target issue.

        Args:
            repo: Name of the repository in format {owner}/{repo}
            number: Issue number
            body: Body of the comment in markdown format
        """
        issue = self._get_issue(repo_name=repo, number=number)
        comment = issue.create_comment(body)

        return "Issue comment created:\n" + json.dumps(
            self._comment_to_json(comment), ensure_ascii=False
        )

    @function
    def add_pull_request_comment(self, repo: str, number: int, body: str):
        """
        Add a comment to the target pull request.

        Args:
            repo: Name of the repository in format {owner}/{repo}
            number: Pull request number
            body: Body of the comment in markdown format
        """
        pr = self._get_pull_request(repo_name=repo, number=number)
        comment = pr.create_issue_comment(body)

        return "Issue comment created:\n" + json.dumps(
            self._comment_to_json(comment), ensure_ascii=False
        )

    @function
    def edit_issue(
        self,
        repo: str,
        number: int,
        title: str = None,
        body: str = None,
        labels: List[str] = None,
        assignees: List[str] = None,
        state: Literal["open", "closed"] = None,
    ):
        """
        Edit an existing issue. You can also close or reopen an issue by specifying the state parameter.

        Args:
            repo: Name of the repository in format {owner}/{repo}
            number: Issue number
            title: Title of the issue
            body: Body of the issue in markdown format
            labels: List of labels to add to the issue
            assignees: List of users to assign to the issue
            state: Whether the issue is open or closed
        """
        issue = self._get_issue(repo_name=repo, number=number)

        issue.edit(
            title=_default_not_set(title),
            body=_default_not_set(body),
            labels=_default_not_set(labels),
            assignees=_default_not_set(assignees),
            state=_default_not_set(state),
        )

        return "Issue updated to:\n" + json.dumps(
            self._issue_pr_to_json(issue), ensure_ascii=False
        )

    @function
    def edit_pull_request(
        self,
        repo: str,
        number: int,
        title: str = None,
        body: str = None,
        base: str = None,
        labels: List[str] = None,
        assignees: List[str] = None,
        state: Literal["open", "closed"] = None,
    ):
        """
        Edit an existing pull request. You can also close or reopen a pull request by specifying the state parameter.

        Args:
            repo: Name of the repository in format {owner}/{repo}
            number: Pull request number
            title: Title of the pull request
            body: Body of the pull request in markdown format
            base: Base branch of the pull request, i.e., the branch with your changes
            labels: List of labels to add to the pull request
            assignees: List of users to assign to the pull request
            state: Whether the pull request is open or closed
        """
        pr = self._get_pull_request(repo_name=repo, number=number)

        pr.edit(
            title=_default_not_set(title),
            body=_default_not_set(body),
            state=_default_not_set(state),
            base=_default_not_set(base),
        )

        if labels:
            pr.add_to_labels(*labels)

        if assignees:
            pr.add_to_assignees(*assignees)

        return "Pull Request updated to:\n" + json.dumps(
            self._issue_pr_to_json(pr), ensure_ascii=False
        )



from .logger import logger
from .to_async_fn import to_async_fn
from .sanitize_schema import sanitize_schema
from .parse_docstring import parse_docstring
from .is_template_str import is_template_str
from ._cloud import is_cloud_env

__all__ = [
    "logger",
    "to_async_fn",
    "sanitize_schema",
    "parse_docstring",
    "is_cloud_env",
    "is_template_str",
]



import string


def is_template_str(s: str) -> bool:
    """
    Check if a string is a template string.
    See: https://stackoverflow.com/a/46161774/2369823

    Args:
        s: A string to check.
    """
    for item in string.Formatter().parse(s):
        if item[1] is not None:
            return True

    return False



import asyncio
import json
import os
from typing import List
from textwrap import dedent

from googleapiclient.errors import HttpError
from markdown import markdown
from simplegmail.message import Message

from npiai import FunctionTool, function
from npiai.error import UnauthorizedError
from npiai.context import Context

from google.oauth2.credentials import Credentials as GoogleCredentials
from oauth2client.client import OAuth2Credentials

from .client import GmailClientWrapper


def convert_google_cred_to_oauth2_cred(
    google_credentials: GoogleCredentials,
) -> OAuth2Credentials:
    # Convert expiry datetime to string if necessary
    expiry = (
        google_credentials.expiry.isoformat() if google_credentials.expiry else None
    )

    # Create an instance of OAuth2Credentials
    return OAuth2Credentials(
        access_token=google_credentials.token,
        client_id=google_credentials.client_id,
        client_secret=google_credentials.client_secret,
        refresh_token=google_credentials.refresh_token,
        token_expiry=google_credentials.expiry,
        token_uri=google_credentials.token_uri,
        user_agent=None,  # or any user agent you have
        scopes=google_credentials.scopes,
    )


class Gmail(FunctionTool):
    name = "gmail"
    description = 'interact with Gmail using English, e.g., gmail("send an email to test@gmail.com")'
    system_prompt = "You are a Gmail Agent helping users to manage their emails"

    _creds: GoogleCredentials | None
    _gmail_client: GmailClientWrapper

    def __init__(self, creds: GoogleCredentials | None = None):
        super().__init__()
        self._creds = creds

    async def start(self):
        if self._creds is None:
            cred_file = os.environ.get("GOOGLE_CREDENTIAL")
            if cred_file is None:
                raise UnauthorizedError("Google credential file not found")
            self._creds = GoogleCredentials.from_authorized_user_file(
                filename=cred_file, scopes="https://mail.google.com/"
            )

        self._gmail_client = GmailClientWrapper(
            _creds=convert_google_cred_to_oauth2_cred(self._creds)
        )
        await super().start()

    def _get_messages_from_ids(self, message_ids: List[str]) -> List[Message]:
        emails: List[Message] = []

        for message_id in message_ids:
            try:
                emails.append(self._gmail_client.get_message_by_id(message_id))
            except HttpError:
                pass

        return emails

    @staticmethod
    def _message_to_string(message: Message) -> str:
        return (
            dedent(
                f"""
            Message ID: {message.id}
            Thread ID: {message.thread_id}
            Sender ID: {message.headers.get('Message-ID', message.id)}
            From: {message.sender}
            To: {message.recipient}
            Subject: {message.subject}
            """
            )
            + f"Content: {message.plain or message.html}"
        )

    @function
    def add_labels(self, message_ids: List[str], labels: List[str]) -> str:
        """
        Add labels to the related emails in the previous chat.
        If the target email is not provided, you should search for it first.

        Args:
            message_ids: A list of IDs of messages that should be labeled. You can find this in the "Message ID: ..." line of the email.
            labels: A list of labels to add.
        """
        messages = self._get_messages_from_ids(message_ids)

        if len(messages) == 0:
            raise Exception("Error: No messages found for the given IDs")

        label_name_map = {
            label.name: label for label in self._gmail_client.list_labels()
        }
        labels_to_add = []

        for lbl in labels:
            if lbl not in label_name_map:
                new_label = self._gmail_client.create_label(lbl)
                label_name_map[lbl] = new_label
            labels_to_add.append(label_name_map[lbl])

        for msg in messages:
            self._gmail_client.add_labels(msg, labels_to_add)

        return "Labels added"

    @function
    def remove_labels(self, message_ids: List[str], labels: List[str]) -> str:
        """
        Remove labels from the related emails in the previous chat.
        If the target email is not provided, you should search for it first.

        Args:
            message_ids: A list of IDs of messages that should be labeled. You can find this in the "Message ID: ..." line of the email.
            labels: A list of labels to remove.
        """
        messages = self._get_messages_from_ids(message_ids)

        if len(messages) == 0:
            raise Exception("Error: No messages found for the given IDs")

        label_name_map = {
            label.name: label for label in self._gmail_client.list_labels()
        }
        labels_to_remove = []

        for lbl in labels:
            if lbl in label_name_map:
                labels_to_remove.append(label_name_map[lbl])

        if len(labels_to_remove):
            for msg in messages:
                self._gmail_client.remove_labels(msg, labels_to_remove)

        return "Labels removed"

    @function
    def create_draft(
        self,
        to: str,
        subject: str,
        message: str = None,
        cc: List[str] = None,
        bcc: List[str] = None,
    ) -> str:
        """
        Create an email draft.

        Args:
            to: The email address the message being sent to.
            subject: The subject line of the email.
            message: The email content in Markdown format.
            cc: The list of email addresses to cc.
            bcc: The list of email addresses to bcc.
        """
        msg = self._gmail_client.create_draft(
            sender="",
            to=to,
            cc=cc,
            bcc=bcc,
            subject=subject,
            msg_plain=message,
            msg_html=markdown(message),
        )

        return "The following draft is created:\n" + self._message_to_string(msg)

    @function
    def create_reply_draft(
        self,
        to: str,
        subject: str,
        recipient_id: str,
        message: str = None,
        cc: List[str] = None,
        bcc: List[str] = None,
    ):
        """
        Create a draft that replies to the last email retrieved in the previous chat.
        If the target email is not provided, you should search for it first.

        Args:
            to: The email address the message being sent to.
            subject: The subject line of the email.
            recipient_id: The ID of the recipient being replied to. You can find this in the "Sender ID: ..." line of the email.
            message: The reply message in Markdown format. You should also quote the email being replied to.
                For example, given the following message:
                    Message ID: 18eaefa2c6b35409
                    Thread ID: 18eaf09f637a31a9
                    Sender ID: <qbvDktEYR9KqlLxxrp6s-w@geopod-ismtpd-15>
                    From: "Test" <test@gmail.com>
                    To: "Me" <me@gmail.com>
                    Date: 2024-04-05 23:30:27+09:00
                    Subject: Greeting
                    Content: Hello there!
                You should generate a reply message like:
                    Hey! How can I help you?

                    On Apr 5, 2024 at 23:30:27, Test <test@gmail.com> wrote:
                    > Hello there!
            cc: The list of email addresses to cc.
            bcc: The list of email addresses to bcc.
        """
        msg = self._gmail_client.create_draft(
            sender="",
            to=to,
            cc=cc,
            bcc=bcc,
            subject=subject,
            msg_plain=message,
            msg_html=markdown(message),
            reply_to=recipient_id,
        )

        return "The following reply draft is created:\n" + self._message_to_string(msg)

    @function
    def reply(
        self,
        to: str,
        subject: str,
        recipient_id: str,
        message: str = None,
        cc: List[str] = None,
        bcc: List[str] = None,
    ):
        """
        Reply to the last email retrieved in the previous chat.
        If the target email is not provided, you should search for it first.

        Args:
            to: The email address the message being sent to.
            subject: The subject line of the email.
            recipient_id: The ID of the recipient being replied to. You can find this in the "Sender ID: ..." line of the email.
            message: The reply message in Markdown format. You should also quote the email being replied to.
                For example, given the following message:
                    Message ID: 18eaefa2c6b35409
                    Thread ID: 18eaf09f637a31a9
                    Sender ID: <qbvDktEYR9KqlLxxrp6s-w@geopod-ismtpd-15>
                    From: "Test" <test@gmail.com>
                    To: "Me" <me@gmail.com>
                    Date: 2024-04-05 23:30:27+09:00
                    Subject: Greeting
                    Content: Hello there!
                You should generate a reply message like:
                    Hey! How can I help you?

                    On Apr 5, 2024 at 23:30:27, Test <test@gmail.com> wrote:
                    > Hello there!
            cc: The list of email addresses to cc.
            bcc: The list of email addresses to bcc.
        """
        msg = self._gmail_client.send_message(
            sender="",
            to=to,
            cc=cc,
            bcc=bcc,
            subject=subject,
            msg_plain=message,
            msg_html=markdown(message),
            reply_to=recipient_id,
        )

        return "The following reply is sent:\n" + self._message_to_string(msg)

    @function
    def search_emails(self, query: str = None, max_results: int = 100) -> str:
        """
        Search for emails with a query.

        Args:
            query: A Gmail query to match emails.
            max_results: Maximum number of results to return.
        """
        msgs = self._gmail_client.get_messages(
            query=query,
            max_results=max_results,
        )

        return json.dumps(
            [self._message_to_string(m) for m in msgs], ensure_ascii=False
        )

    @function
    async def send_email(
        self,
        ctx: Context,
        to: str,
        subject: str,
        message: str = None,
        cc: List[str] = None,
        bcc: List[str] = None,
    ):
        """
        Send an email.

        Args:
            ctx: NPi context
            to: The email address the message being sent to.
            subject: The subject line of the email.
            message: The email content in Markdown format.
            cc: The list of email addresses to cc.
            bcc: The list of email addresses to bcc.
        """
        # approved = await self.hitl.confirm(
        #     ctx=ctx,
        #     tool_name=self.name,
        #     message=f'The following email will be sent to {to}:\n {message}'
        # )
        #
        # if not approved:
        #     return 'The email could not be sent due to user rejection'

        msg = self._gmail_client.send_message(
            sender="",
            to=to,
            cc=cc,
            bcc=bcc,
            subject=subject,
            msg_plain=message,
            msg_html=markdown(message),
        )

        return "Sending Success\n" + self._message_to_string(msg)

    @function
    async def wait_for_reply(self, sender: str):
        """
        Wait for reply from the last email sent in the previous chats.

        Args:
            sender: The reply message sender. This should be the recipient of the last email, and you can find it in the "To: ..." line.
        """
        while True:
            messages = self._gmail_client.get_messages(
                query=f"is:unread from:{sender}",
                max_results=1,
            )

            if len(messages):
                msg = messages[0]
                msg.mark_as_read()
                return self._message_to_string(msg)

            await asyncio.sleep(3)



import asyncio
from npiai import agent
from npiai.tools.web import Twitter


async def main():
    async with agent.wrap(Twitter(headless=False)) as twitter:
        return await twitter.chat(
            "Find the latest 10 tweets by @AtomSilverman. You should also include the link to the tweet. You must try to load more tweets until exactly 10 tweets are found. Output all tweets in your final response."
        )


if __name__ == "__main__":
    asyncio.run(main())



from npiai.core.tool import (
    FunctionTool,
    BrowserTool,
    function,
    AgentTool,
    BrowserAgentTool,
)
from npiai.core.hitl import HITL
from npiai.llm import OpenAI, LLM
from npiai.core.base import Context
from npiai.types import FromContext

__all__ = [
    "BrowserTool",
    "FunctionTool",
    "AgentTool",
    "BrowserAgentTool",
    "function",
    "HITL",
    "OpenAI",
    "LLM",
    "Context",
    "FromContext",
]



"""This module contains the classes for the context and context message"""

import datetime
import json
import uuid
import asyncio
from typing import List, Union, Dict, TYPE_CHECKING, TypeVar, Type, Any
from textwrap import dedent

from mem0 import Memory
from litellm.types.completion import ChatCompletionMessageParam
from pydantic import create_model, Field

from npiai.utils import sanitize_schema, logger

if TYPE_CHECKING:
    from npiai.core import BaseTool


_T = TypeVar("_T")


class Task:
    """the message wrapper of a message in the context"""

    def __init__(self, goal: str) -> None:
        self.task_id = str(uuid.uuid4())
        self.goal = goal
        self.born_at = datetime.datetime.now()
        self.dialogues: List[Union[ChatCompletionMessageParam]] = []
        self.response: str
        self._session: Union["Context", None] = None

    async def step(self, msgs: List[Union[ChatCompletionMessageParam]]) -> None:
        """add a message to the context"""
        self.dialogues.extend(msgs)

        # for msg in msgs:
        #     if msg.get('content'):
        #         await self._session.send(callback.Callable(msg.get('content')))

    def set_session(self, session: "Context") -> None:
        self._session = session

    def conversations(self) -> List[ChatCompletionMessageParam]:
        """return the raw message"""
        return self.dialogues.copy()


class Context:
    id: str

    _q: asyncio.Queue
    _is_finished: bool
    _is_failed: bool
    _result: str
    _failed_msg: str
    _last_screenshot: str | None
    _active_tool: Union["BaseTool", None]
    _memory: Memory
    _query_cache: Dict[str, Any]

    def __init__(
        self,
        memory: Memory = Memory(),
    ) -> None:
        self.id = str(uuid.uuid4())
        self._q = asyncio.Queue()
        self._is_finished = False
        self._result = ""
        self._is_failed = False
        self._failed_msg = ""
        self._last_screenshot = None
        self._active_tool = None
        self._memory = memory
        self._query_cache = {}

    async def _ask_human(self, query: str):
        """
        Ask human if no memory is found

        Args:
            query: Memory search query
        """
        if not self._active_tool:
            raise RuntimeError("No active tool found")

        res = await self._active_tool.hitl.input(
            ctx=self,
            tool_name=self._active_tool.name,
            message=f"Please provide the following information: {query}",
        )

        await self.save(f"Question: {query}? Answer: {res}")

    async def save(self, info: str):
        """
        Save the given information into memory

        Args:
            info: Information to save
        """
        m = self._memory.add(
            data=info,
            run_id=self.id,
            metadata={"raw": info},
            prompt=dedent(
                f"""
                Deduce the facts, preferences, and memories from the provided text.
                Just return the facts, preferences, and memories in bullet points:
                Natural language text: {info}
                
                Constraint for deducing facts, preferences, and memories:
                - The facts, preferences, and memories should be concise and informative.
                - Don't start by "The person likes Pizza". Instead, start with "Likes Pizza".
                - Don't remember the user/agent details provided. Only remember the facts, preferences, and memories.
                
                Deduced facts, preferences, and memories:
                """
            ),
        )
        # clear cache
        self._query_cache = {}
        logger.debug(f"Saved memory: {m}")

    async def ask(
        self,
        query: str,
        return_type: Type[_T] = str,
        constraints: str = None,
        _is_retry: bool = False,
    ) -> _T:
        """
        Search the memory

        Args:
            query: Memory search query
            return_type: Return type of the result
            constraints: Search constraints
            _is_retry: Retry flag
        """
        cached_result = self._query_cache.get(query, None)

        if cached_result:
            return cached_result

        async def retry():
            if _is_retry:
                return

            # invoke HITL and retry
            await self._ask_human(query)
            return await self.ask(query, return_type, constraints, _is_retry=True)

        memories = self._memory.search(query, run_id=self.id, limit=10)
        logger.debug(f"Retrieved memories: {json.dumps(memories)}")

        if len(memories) == 0:
            logger.info(f"No memories found for query: {query}")
            return await retry()

        mem_str = "- " + "\n- ".join(
            "Extracted data: " + m["text"] + "\n Raw data: " + m["metadata"]["raw"]
            for m in memories
        )

        callback_model = create_model(
            "MemorySearchCallback",
            data=(
                return_type,
                Field(
                    default=None,
                    description="Retrieved memories. Set to `null` if no data is found.",
                ),
            ),
        )

        schema = sanitize_schema(callback_model)

        # TODO: use npi llm client?
        response = self._memory.llm.generate_response(
            tool_choice="required",
            messages=[
                {
                    "role": "system",
                    "content": dedent(
                        f"""
                        Optimize the functioning of a memory retrieval AI tool by adhering to the following guidelines for processing search queries:

                        1. Respond to search queries by invoking the `callback` function, supplying the required information within the `data` argument.
                        2. When delivering the `data`, aim for brevity and accuracy—only include the core details pertinent to the query.
                           - For instance, rather than "The capital of France is Paris," provide the succinct "Paris."
                        3. Ensure the content in `data` is an exact match to the search query.
                           - As an example, "The password for the Wi-Fi is '1234Abcd'" should only be returned if the query is "Wi-Fi password" and not for "macOS password" or any other password-related question.
                        4. Adhere to any specified constraints in the retrieval process.

                        ## Memories
                        """
                    )
                    + mem_str,
                },
                {
                    "role": "user",
                    "content": dedent(
                        f"""
                        Query: {query}
                        Constraints: {constraints}
                        """
                    ),
                },
            ],
            tools=[
                {
                    "type": "function",
                    "function": {
                        "name": "callback",
                        "description": "Callback with retrieved information from the given memory",
                        "parameters": schema,
                    },
                }
            ],
        )

        tool_calls = response["tool_calls"]

        if not tool_calls or tool_calls[0]["name"] != "callback":
            logger.info(
                f"No LLM callback found for query: {query}. Response: {json.dumps(response)}"
            )
            return await retry()

        logger.debug(f"LLM callback: {json.dumps(tool_calls)}")

        data = tool_calls[0]["arguments"].get("data", None)

        if data is None:
            logger.info(f"No data found for query: {query}")
            return await retry()

        self._query_cache[query] = data

        return data

    # @abstractmethod
    # NOTE: this method should not be abstract
    # since we need to create an empty context when running locally
    # SEE: FunctionTool::call()
    def credentials(self, app_code: str) -> Dict[str, str]:
        return {}

    def entry(self):
        pass

    def exit(self):
        pass

    def with_task(self, task: Task) -> "Context":
        task.set_session(self)
        return self

    def fork(self, task: Task) -> "Context":
        """fork a child message, typically when call a new tools"""
        return self.with_task(task)

    async def refresh_screenshot(self) -> str | None:
        """
        Refresh and return the latest screenshot of the running tools.

        Returns:
            None if no screenshot is available or the screenshot stays unchanged.
            Otherwise, return the latest screenshot.
        """
        if not self._active_tool or self.is_finished():
            return None

        screenshot = await self._active_tool.get_screenshot()

        if screenshot == self._last_screenshot:
            return None

        self._last_screenshot = screenshot

        return screenshot

    async def send(self, cb: str) -> None:
        """send a message to the context"""
        # self.cb_dict[cb.id()] = cb
        await self._q.put(cb)

    async def fetch(self):
        """receive a message"""
        while not self.is_failed() and not self.is_finished():
            try:
                item = self._q.get_nowait()
                if item:
                    return item
            except asyncio.QueueEmpty:
                await asyncio.sleep(0.01)

    def get_callback(self, cb_id: str):
        pass
        # return self.cb_dict[cb_id]

    def finish(self, msg: str):
        self._result = msg
        self._is_finished = True
        self._last_screenshot = None
        self._q.task_done()

    def failed(self, msg: str):
        self._failed_msg = msg
        self._is_failed = True
        self._q.task_done()

    def is_finished(self) -> bool:
        return self._is_finished

    def get_result(self) -> str:
        return self._result

    def is_failed(self) -> bool:
        return self._is_failed

    def get_failed_msg(self) -> str:
        return self._failed_msg

    def bind(self, tool: "BaseTool"):
        self._active_tool = tool

    def get_tool(self) -> Union["BaseTool", None]:
        return self._active_tool



import asyncio

from npiai import agent
from npiai.tools.google import Gmail
from google.oauth2.credentials import Credentials


async def main():
    credentials = Credentials.from_authorized_user_file("gmail_token.json")

    async with agent.wrap(Gmail(credentials)) as gmail:
        return await gmail.chat(
            'Add label "TEST" to the latest email from daofeng.wu@emory.edu'
        )


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import json
import os
from typing import Literal

from slack_sdk.web.async_client import AsyncWebClient, AsyncSlackResponse

from npiai.context import Context
from npiai.utils import logger
from npiai import FunctionTool, function
from npiai.error.auth import UnauthorizedError

__PROMPT__ = """
You are a Slack Agent helping user send/retrieve messages to/from discord channels. 

For any tool requesting a discord channel ID, you must search for it in the chat history first. 
If not found, you should ask the user to enter the ID.

## Example

Task: Send a DM to user {{user_id}} and wait for their reply.
Workflow:
- Create a DM channel with user {{user_id}}. Record the DM channel ID {{channel_id}}.
- Send a message in the DM channel. Record the message ID {{message_id}}.
- Wait for the reply for message {{message_id}} to appear in channel {{channel_id}}.
"""


class Slack(FunctionTool):
    name = "slack"
    description = "Send/Retrieve messages to/from Slack channels"
    system_prompt = __PROMPT__

    _access_token: str | None
    _client: AsyncWebClient
    _user_id: str = None

    def __init__(self, access_token: str = None):
        super().__init__()
        self._access_token = access_token or os.environ.get("SLACK_ACCESS_TOKEN", None)

    async def start(self):
        if self._access_token is None:
            raise UnauthorizedError("Slack credentials are not found")

        self._client = AsyncWebClient(token=self._access_token)

    async def _get_user_id(self) -> str:
        if not self._user_id:
            res = await self._client.auth_test()
            self._user_id = res["user_id"]

        return self._user_id

    @staticmethod
    def _parse_raw_message(raw_message: dict):
        return {
            "thread_id": raw_message["ts"],
            "user": raw_message["user"],
            "content": raw_message["text"],
            "reply_count": raw_message.get("reply_count", 0),
        }

    def _get_messages_from_response(self, response: AsyncSlackResponse):
        messages = []

        for msg in response["messages"]:
            if msg["type"] != "message":
                continue
            messages.append(self._parse_raw_message(msg))

        return sorted(messages, key=lambda x: float(x["thread_id"]))

    @function
    async def ask_for_id(
        self, ctx: Context, name: Literal["user", "channel", "thread"]
    ):
        """
        Ask the user to provide recipient's user id, channel id, or thread id.
        Args:
            ctx: NPi context.
            name: The type of id to ask for.
        """
        return await self.hitl.input(
            ctx,
            self.name,
            f"Please provide recipient's {name} ID to send messages to",
        )

    @function
    async def list_channels(self):
        """Get a list of all Slack channels"""
        res = await self._client.conversations_list()
        channels = []

        for chn in res["channels"]:
            channels.append(
                {
                    "id": chn["id"],
                    "name": chn["name"],
                }
            )

        return json.dumps(channels, ensure_ascii=False)

    @function
    async def create_dm(self, user_id: str):
        """
        Create a direct message channel with a specific user.

        Args:
            user_id: The ID of the **user** who will receive the direct message. Note that this is not the channel ID.
        """
        res = await self._client.conversations_open(users=[user_id])

        return f'Direct message channel created. Channel ID: {res["channel"]["id"]}'

    @function
    async def send_message(self, channel_id: str, message: str):
        """
        Send a message to the Slack channel with the given channel ID.

        Args:
            channel_id: The ID of the **channel** to send the message to. You should ask the user for it if not provided.
            message: The message to send.
        """
        res = await self._client.chat_postMessage(
            channel=channel_id,
            text=message,
        )

        msg = res["message"]

        logger.debug(f'[{self.name}]: Sent message: (id: {msg["ts"]}) {msg["text"]}')

        return f'Message sent. Thread ID: {msg["ts"]}'

    @function
    async def fetch_history(self, channel_id: str, max_messages: int = 1):
        """
        Fetch history messages from the Slack channel with the given channel ID.

        Args:
            channel_id: The ID of the **channel** to fetch the history from.
            max_messages: The maximum number of messages to fetch.
        """
        res = await self._client.conversations_history(
            channel=channel_id, limit=max_messages
        )
        messages = self._get_messages_from_response(res)

        logger.debug(
            f"[{self.name}]: Fetched {len(messages)} messages: {json.dumps(messages, indent=2, ensure_ascii=False)}"
        )

        return json.dumps(messages, ensure_ascii=False)

    @function
    async def reply(self, channel_id: str, thread_id: str, message: str):
        """
        Reply to a context in the Slack channel with the given channel ID.

        Args:
            channel_id: The ID of the **channel** to reply to.
            thread_id: The ID of the **context** to reply to.
            message: The message to reply.
        """
        res = await self._client.chat_postMessage(
            channel=channel_id,
            text=message,
            thread_ts=thread_id,
        )

        msg = res["message"]

        logger.debug(
            f'[{self.name}]: Created reply for context ID {thread_id}. Reply: (id: {msg["ts"]}) {msg["text"]}'
        )

        return f'Created reply for message ID {msg.id}. Reply ID: {msg["ts"]}'

    @function
    async def wait_for_reply(self, channel_id: str, thread_id: str):
        """
        Wait for a reply to the given context in the Slack channel with the given channel ID.

        Args:
            channel_id: The ID of the **channel** to wait for.
            thread_id: The ID of the **context** to wait for.
        """
        last_ts = thread_id

        while True:
            # check threaded replies
            thread_res = await self._client.conversations_replies(
                channel=channel_id,
                ts=thread_id,
            )

            if len(thread_res["messages"]) > 1:
                messages = self._get_messages_from_response(thread_res)[1:]

                logger.debug(
                    f"[{self.name}]: Found {len(messages)} context replies: {json.dumps(messages, indent=2, ensure_ascii=False)}"
                )

                return json.dumps(messages, ensure_ascii=False)

            # check channel replies
            channel_res = await self._client.conversations_history(
                channel=channel_id,
                oldest=last_ts,
                inclusive=False,
                limit=10,
            )

            for msg in channel_res["messages"]:
                if (
                    msg["type"] == "message"
                    and msg["user"] != await self._get_user_id()
                ):
                    messages = [self._parse_raw_message(msg)]

                    logger.debug(
                        f"[{self.name}]: Received a new message: {json.dumps(messages, indent=2, ensure_ascii=False)}"
                    )

                    return json.dumps(messages, ensure_ascii=False)

            await asyncio.sleep(3)



import asyncio

from npiai import agent
from npiai.tools import Discord
from npiai.hitl_handler import ConsoleHandler


async def main():
    async with agent.wrap(Discord()) as agent_tool:
        agent_tool.use_hitl(ConsoleHandler())
        return await agent_tool.chat(
            "Send a direct message to Dolphin asking if he is doing well, and wait for his reply."
        )


if __name__ == "__main__":
    asyncio.run(main())



from abc import ABC, abstractmethod
from typing import List, Dict, Any

from litellm.types.completion import ChatCompletionToolMessageParam
from litellm.types.utils import ChatCompletionMessageToolCall
from openai.types.chat import ChatCompletionToolParam

from npiai.context import Context
from npiai.core.hitl import HITL
from npiai.types import FunctionRegistration
from npiai.constant import CTX_QUERY_POSTFIX


class BaseTool(ABC):
    name: str = ""
    description: str = ""
    system_prompt: str = ""
    provider: str = "private"

    _fn_map: Dict[str, FunctionRegistration]
    _hitl: HITL | None

    @classmethod
    def from_context(cls, ctx: Context) -> "BaseTool":
        # bind the tool to the Context
        raise NotImplementedError(
            "subclasses must implement this method for npi cloud hosting"
        )

    @property
    def hitl(self) -> HITL:
        if self._hitl is None:
            raise AttributeError("HITL handler has not been set")

        return self._hitl

    @property
    def tools(self) -> List[ChatCompletionToolParam]:
        fns = self.unpack_functions()
        result = []
        for fn in fns:
            result.append(fn.get_tool_param())
        return result

    @abstractmethod
    def unpack_functions(self) -> List[FunctionRegistration]:
        """Export the functions registered in the tools"""
        ...

    @abstractmethod
    async def start(self):
        """Start the tools"""
        ...

    @abstractmethod
    async def end(self):
        """Stop and dispose the tools"""
        ...

    async def exec(self, ctx: Context, fn_name: str, args: Dict[str, Any] = None):
        if fn_name not in self._fn_map:
            raise RuntimeError(
                f"[{self.name}]: function `{fn_name}` not found. Available functions: {self._fn_map.keys()}"
            )

        if not args:
            args = {}

        fn = self._fn_map[fn_name]

        # add context param
        if fn.ctx_param_name is not None:
            args[fn.ctx_param_name] = ctx

        # add context variables
        for ctx_var in fn.ctx_variables:
            query = args.pop(f"{ctx_var.name}{CTX_QUERY_POSTFIX}", ctx_var.query)

            args[ctx_var.name] = await ctx.ask(
                query=query,
                return_type=ctx_var.return_type,
                constraints=ctx_var.constraints,
            )

        return await fn.fn(**args)

    def use_hitl(self, hitl: HITL):
        self._hitl = hitl

    def schema(self):
        """
        Find the wrapped tool functions and export them as yaml
        """
        return {
            "kind": "Function",
            "metadata": {
                "name": self.name,
                "description": self.description,
                "provider": self.provider,
            },
            "spec": {
                "runtime": {
                    "language": "python",
                    "version": "3.11",
                },
                "dependencies": [
                    {
                        "name": "npiai",
                        "version": "0.1.0",
                    }
                ],
                "functions": [t.get_meta() for t in self.unpack_functions()],
            },
        }

    async def get_screenshot(self) -> str | None:
        return None

    # Context manager
    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.end()


class BaseFunctionTool(BaseTool, ABC):
    @abstractmethod
    async def call(
        self,
        tool_calls: List[ChatCompletionMessageToolCall],
        ctx: Context | None = None,
    ) -> List[ChatCompletionToolMessageParam]: ...


class BaseAgentTool(BaseTool, ABC):
    @abstractmethod
    async def chat(
        self,
        ctx: Context,
        instruction: str,
    ) -> str: ...



import base64

from markdownify import MarkdownConverter
from playwright.async_api import ElementHandle, Error

from npiai.core.browser import PlaywrightContext
from npiai.utils import logger

from ._function import FunctionTool, function


class MdConverter(MarkdownConverter):
    # skip <noscript> tags
    def convert_noscript(self, _el, _text, _convert_as_inline):
        return ""


class BrowserTool(FunctionTool):
    use_screenshot: bool
    playwright: PlaywrightContext

    def __init__(
        self,
        playwright: PlaywrightContext = None,
        use_screenshot: bool = True,
        headless: bool = True,
    ):
        """
        Initialize a Browser App

        Args:
            playwright: Playwright context to use. A new playwright context is created if None.
            use_screenshot: Whether to send a screenshot of the current page to the vision model. This should be used with a navigator and a vision model.
            headless: Whether to run playwright in headless mode.
        """
        super().__init__()
        self.use_screenshot = use_screenshot
        self.playwright = playwright or PlaywrightContext(headless)

    @function
    async def get_text(self):
        """Get the text content (as markdown) of the current page"""
        html = await self.playwright.page.evaluate("() => document.body.innerHTML")
        return MdConverter().convert(html)

    async def start(self):
        """Start the Browser App"""
        if not self._started:
            await super().start()
            await self.playwright.start()

    async def end(self):
        """Dispose the chrome tools"""
        await super().end()
        await self.playwright.stop()

    async def goto_blank(self):
        """
        Go to about:blank page
        This can be used as a cleanup function when a context finishes
        """
        await self.playwright.page.goto("about:blank")

    async def get_screenshot(self) -> str | None:
        """Get the screenshot of the current page"""
        if (
            not self.playwright
            or not self.playwright.ready
            or self.playwright.page.url == "about:blank"
        ):
            return None

        try:
            screenshot = await self.playwright.page.screenshot(caret="initial")
            return "data:image/png;base64," + base64.b64encode(screenshot).decode()
        except Error as e:
            logger.error(e)
            return None

    async def get_page_url(self):
        """Get the URL of the current page"""
        return self.playwright.page.url

    async def get_page_title(self):
        """Get the page title of the current page"""
        return await self.playwright.page.title()

    async def is_scrollable(self):
        """Check if the current page is scrollable"""
        return await self.playwright.page.evaluate("() => npi.isScrollable()")

    async def get_interactive_elements(self, screenshot: str):
        """Get the interactive elements of the current page"""
        return await self.playwright.page.evaluate(
            """async (screenshot) => {
                const { elementsAsJSON, addedIDs } = await npi.snapshot(screenshot);
                return [elementsAsJSON, addedIDs];
            }""",
            screenshot,
        )

    async def get_element_by_marker_id(self, elem_id: str):
        """Get the element by marker id"""
        handle = await self.playwright.page.evaluate_handle(
            "id => npi.getElement(id)",
            elem_id,
        )

        elem_handle = handle.as_element()

        if not elem_handle:
            raise Exception(f"Element not found (id: {elem_id})")

        return elem_handle

    async def element_to_json(self, elem: ElementHandle):
        """
        Convert the element into JSON

        Args:
            elem: Playwright element handle

        Returns:
            JSON representation of the element
        """
        return await self.playwright.page.evaluate(
            "(elem) => npi.elementToJSON(elem)",
            elem,
        )

    async def init_observer(self):
        """Initialize a mutation observer on the current page"""
        await self.playwright.page.evaluate("() => npi.initObserver()")

    async def wait_for_stable(self):
        """Wait for the current page to be stable"""
        await self.playwright.page.evaluate("() => npi.stable()")

    async def add_bboxes(self):
        """Add bounding boxes to the interactive elements on the current page"""
        await self.playwright.page.evaluate("() => npi.addBboxes()")

    async def clear_bboxes(self):
        """Clear the bounding boxes on the current page"""
        await self.playwright.page.evaluate("() => npi.clearBboxes()")

    async def click(self, elem: ElementHandle):
        """
        Click an element on the page

        Args:
            elem: Element Handle of the target element
        """

        try:
            await elem.click()
        except TimeoutError:
            await self.playwright.page.evaluate(
                "(elem) => npi.click(elem)",
                elem,
            )

        return f"Successfully clicked element"

    async def fill(self, elem: ElementHandle, value: str):
        """
        Fill in an input field on the page

        Args:
            elem: Element Handle of the target element
            value: value to fill
        """

        try:
            await elem.fill(value)
        except TimeoutError:
            await self.playwright.page.evaluate(
                "([elem, value]) => npi.fill(elem, value)",
                [elem, value],
            )

        return f"Successfully filled value {value} into element"

    async def select(self, elem: ElementHandle, value: str):
        """
        Select an option for a <select> element

        Args:
            elem: Element Handle of the target element
            value: value to select
        """

        try:
            await elem.select_option(value)
        except TimeoutError:
            await self.playwright.page.evaluate(
                "([elem, value]) => npi.select(elem, value)",
                [elem, value],
            )

        return f"Successfully selected value {value} for element"

    async def enter(self, elem: ElementHandle):
        """
        Press Enter on an input field. This action usually submits a form.

        Args:
            elem: Element Handle of the target element
        """

        try:
            await elem.press("Enter")
        except TimeoutError:
            await self.playwright.page.evaluate(
                "(elem) => npi.enter(elem)",
                elem,
            )

        return f"Successfully pressed Enter on element"

    async def scroll(self):
        """
        Scroll the page down to reveal more contents
        """

        await self.playwright.page.evaluate("() => npi.scrollPageDown()")
        await self.playwright.page.wait_for_timeout(300)

        return f"Successfully scrolled down to reveal more contents"

    async def back_to_top(self):
        """
        Scroll the page back to the top to start over
        """

        await self.playwright.page.evaluate("() => window.scrollTo(0, 0)")
        await self.playwright.page.wait_for_timeout(300)

        return f"Successfully scrolled to top"



import asyncio
from npiai import agent
from npiai.tools.web import Twitter


async def main():
    async with agent.wrap(Twitter(headless=False)) as twitter:
        return await twitter.chat(
            "Reply to the latest tweet of @elonmusk with your opinions. You can write anything you like.",
        )


if __name__ == "__main__":
    asyncio.run(main())



"""The definition of Google Calendar App"""

import datetime
import json
import os

from googleapiclient.discovery import build, Resource
from google.oauth2.credentials import Credentials as GoogleCredentials

from npiai import function, FunctionTool
from npiai.error import UnauthorizedError
from npiai.context import Context


# https://developers.google.com/calendar/quickstart/python
# API Reference: https://developers.google.com/calendar/api/v3/reference


class GoogleCalendar(FunctionTool):
    name = "google_calendar"
    description = "Manage events on Google Calendar"
    system_prompt = """
    You are an assistant interacting with Google Calendar API. 
    Your job is the selecting the best function based the tool list.
    """

    _creds: GoogleCredentials | None
    _service: Resource

    def __init__(self, creds: GoogleCredentials | None = None):
        super().__init__()
        self._creds = creds

    async def start(self):
        if self._creds is None:
            cred_file = os.environ.get("GOOGLE_CREDENTIAL")
            if cred_file is None:
                raise UnauthorizedError("Google credential file not found")
            self._creds = GoogleCredentials.from_authorized_user_file(
                filename=cred_file, scopes="https://www.googleapis.com/auth/calendar"
            )
        self._service = build("calendar", "v3", credentials=self._creds)
        await super().start()

    @function
    async def get_user_email(self, ctx: Context, message: str) -> str:
        """
        Get the user's email address

        Args:
            ctx: NPi context
            message: The message to ask the user for their email address
        """
        return await self.hitl.input(ctx, self.name, message)

    @function
    def get_today(self):
        """Get today's date"""
        return datetime.date.today().strftime("%a, %Y-%m-%d")

    @function
    def get_timezone(self):
        """Get the user's timezone"""
        res = (
            self._service.calendars()
            .get(calendarId="primary")  # pylint: disable=maybe-no-member
            .execute()
        )

        return res.get("timeZone")

    @function
    def retrieve_events(
        self, calendar_id: str = "primary", time_min: str = None, time_max: str = None
    ) -> str:
        """
        Retrieve events from Google Calendar.

        Args:
            calendar_id: The calendar ID which typically in the form of an email address. The primary calendar will be used if not given.
            time_min: Lower bound (exclusive) for an event's end time to filter by. Must be an RFC3339 timestamp with mandatory time zone offset, for example, 2011-06-03T10:00:00-05:00. If `time_min` is set, `time_max` must be greater than `time_min`.
            time_max: Upper bound (exclusive) for an event's start time to filter by. Must be an RFC3339 timestamp with mandatory time zone offset, for example, 2011-06-05T10:00:00-05:00. If `time_max` is set, `time_min` must be smaller than `time_min`.
        """
        calendar_id = calendar_id
        time_min = time_min
        time_max = time_max
        max_result = 10
        single_events = True
        order_by = "startTime"
        event_types = "default"

        if time_min is None:
            # 'Z' indicates UTC time
            time_min = datetime.datetime.utcnow().strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )  # pylint: disable=maybe-no-member

        events_result = (
            self._service.events()
            .list(  # pylint: disable=maybe-no-member
                calendarId=calendar_id,
                timeMin=time_min,
                timeMax=time_max,
                maxResults=max_result,
                singleEvents=single_events,
                orderBy=order_by,
                eventTypes=event_types,
            )
            .execute()
        )
        events = events_result.get("items", [])

        if not events:
            return "No events found."

        return json.dumps(events, ensure_ascii=False)

    @function
    def create_event(
        self,
        summary: str,
        start_time: str,
        end_time: str,
        calendar_id: str = "primary",
        description: str = "",
    ) -> str:
        """
        Create and add an event to Google Calendar.

        Args:
            summary: The summary of the event.
            start_time: The start time of this event. Must be an RFC3339 timestamp with mandatory time zone offset, for example, 2011-06-03T09:00:00-05:00. You should retrieve the user's timezone first.
            end_time: The end time of this event. Must be an RFC3339 timestamp with mandatory time zone offset, for example, 2011-06-03T10:00:00-05:00. You should retrieve the user's timezone first.
            calendar_id: The calendar ID which typically in the form of an email address. The primary calendar will be used if not given.
            description: The detailed description of the event.
        """
        attendee = ([],)
        location = (None,)
        recurrence = (None,)
        event = {
            "summary": summary,
            "location": location,
            "description": description,
            "start": {
                "dateTime": start_time,
            },
            "end": {
                "dateTime": end_time,
            },
            "recurrence": [recurrence],
            "attendees": attendee,
            "reminders": {
                "useDefault": False,
                "overrides": [
                    {"method": "email", "minutes": 24 * 60},
                    {"method": "popup", "minutes": 10},
                ],
            },
        }

        event = (
            self._service.events()
            .insert(  # pylint: disable=maybe-no-member
                calendarId=calendar_id, body=event
            )
            .execute()
        )

        return f'Event created: {event.get("htmlLink")}'



import base64
import json
from typing import Dict

import requests


class Client:
    __API_VERSION = "2024-07-25"

    def __init__(self, access_token, endpoint="https://api.npi.ai"):
        self._base_url = endpoint
        self._api_version = "2024-07-25"
        self._access_token = (access_token,)
        self.client = requests.Session()
        self.client.headers.update({"Content-Type": "application/json"})
        self.client.headers.update({"Accept": "application/json"})
        self.client.headers.update({"Authorization": f"Bearer {access_token}"})
        self.client.headers.update({"x-npi-api-version": self._api_version})

    @property
    def api_version(self):
        return self._api_version

    def get_credentials(self, app: str) -> Dict[str, str]:
        response = self.client.request("GET", f"{self._base_url}/credentials/{app}")
        if response.status_code != 200:
            raise Exception(f"Failed to get credentials: {response.text}")
        creds = json.loads(response.text)
        # Decoding the Base64 string
        decoded_bytes = base64.b64decode(creds["credentials"])
        return json.loads(decoded_bytes.decode("utf-8"))


if __name__ == "__main__":
    client = Client(
        access_token="eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJNYXBDbGFpbXMiOnsiZXhwIjoxNzIxOTM1MTg1LCJpc3MiOiJucGkuY29tIiwic3ViIjoiZ2l0aHViXzk3NjY4MjEifSwidXVpZCI6IjY2ODhhY2VhMmEwNTgxYTM2ZjQwNjkzOSIsInRvb2xfaWQiOiI2NjlkZjlhMTgxOGNlMmE2OTllOTU2ZmYiLCJjbGllbnRfaWQiOiJGelhBU2pJdVBWMlNXeW5qeHJJVCIsImhpbnQiOiI2NjllMjdlYTAwYmY2MWYzMjQ0ODFiNjUifQ.kh3D7691gwaXNBBudaWu0C46G15rsEOqsZpLEUhXYGc",
        endpoint="http://localhost:8080",
    )
    print(client.get_credentials("github"))



from npiai.types.function_registration import FunctionRegistration, ToolFunction
from npiai.types.shot import Shot
from npiai.types.tool_meta import ToolMeta
from npiai.types.from_context import FromContext

__all__ = ["FunctionRegistration", "ToolFunction", "Shot", "ToolMeta", "FromContext"]



from typing import Callable, Optional, Awaitable, List, Dict, Any, Type
from dataclasses import dataclass, asdict

from openai.types.chat import ChatCompletionToolParam
from pydantic import BaseModel

from npiai.types.shot import Shot
from npiai.types.from_context import FromContext
from npiai.constant import CTX_QUERY_POSTFIX

ToolFunction = Callable[..., Awaitable[str]]


@dataclass(frozen=True)
class FunctionRegistration:
    fn: ToolFunction
    description: str
    name: str
    ctx_variables: List[FromContext]
    ctx_param_name: str | None = None
    schema: Optional[Dict[str, Any]] = None
    model: Optional[Type[BaseModel]] = None
    few_shots: Optional[List[Shot]] = None

    def get_meta(self):
        params = {}

        for name in self.schema:
            # remove context variable queries
            if not name.endswith(CTX_QUERY_POSTFIX):
                params[name] = self.schema[name]

        return {
            "description": self.description,
            "name": self.name,
            "parameters": params,
            "fewShots": (
                [asdict(ex) for ex in self.few_shots] if self.few_shots else None
            ),
        }

    def get_tool_param(self) -> ChatCompletionToolParam:
        tool: ChatCompletionToolParam = {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
            },
        }

        if self.schema is not None:
            tool["function"]["parameters"] = self.schema

        return tool



import os
import asyncio
import sys
import re
import json
import logging
import signal
from typing import List

from fastapi import FastAPI, Request, HTTPException
from fastapi.responses import JSONResponse
import uvicorn

from npiai import utils
from npiai.context import ContextManager
from npiai.core.base import BaseTool

from ._context import CloudContext
from ._client import Client


class ToolRuntime:
    def __init__(
        self,
        tool_cls: List[BaseTool.__class__],
        port: int = 9140,
        endpoint="https://api.npi.ai",
    ):
        self.tools = {}
        self._endpoint = endpoint
        for tool in tool_cls:
            self.tools[tool.name] = tool
        self.port = port
        self.app = FastAPI()
        self.ctx_mgr = ContextManager()

    def run(self):
        self._start()

    def get_tool(self, name: str) -> BaseTool.__class__:
        if name not in self.tools:
            raise ValueError(f"Tool {name} not found")

        return self.tools[name]

    def _start(self):
        """Start the server"""
        if not utils.is_cloud_env():
            raise RuntimeError(
                "Server mode is disabled, if you want to run the server, set env NPIAI_SERVICE_MODE=true"
            )

        def convert_camel_to_snake(name: str) -> str:
            return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()

        @self.app.api_route(
            "/{tool_name}/{full_path:path}",
            methods=["GET", "POST", "PUT", "DELETE", "PATCH"],
        )
        async def root(request: Request, tool_name: str, full_path: str):
            method = convert_camel_to_snake(full_path)
            try:
                ssid = request.headers.get("x-npi-session-id")
                if not ssid:
                    authorization = request.headers.get("authorization")
                    token = authorization.split(" ")
                    access_token = ""
                    if (
                        len(token) == 2 and token[0].lower() == "bearer"
                    ):  # TODO apikey & none auth?
                        access_token = token[1]
                    ctx = CloudContext(
                        req=request,
                        client=Client(
                            access_token=access_token, endpoint=self._endpoint
                        ),
                    )
                    self.ctx_mgr.save_context(ctx)
                else:
                    ctx = self.ctx_mgr.get_context(ssid)
                    if not ctx:
                        raise HTTPException(
                            status_code=404, detail="session context not found"
                        )
                tool_cls = self.get_tool(tool_name)
                async with tool_cls.from_context(ctx) as instance:
                    match request.method:
                        case "POST":
                            args = await request.json()
                            res = await instance.exec(ctx, method, args)
                        case "GET":
                            args = {k: v for k, v in request.query_params.items()}
                            res = await instance.exec(ctx, method, args)
                        case _:
                            return JSONResponse(
                                {"error": "Method not allowed"}, status_code=405
                            )
                    try:
                        return JSONResponse(content=json.loads(res))
                    except json.JSONDecodeError as e:
                        return res
            except Exception as e:
                logging.error(f"Failed to process request: {e}", exc_info=True)
                raise HTTPException(status_code=500, detail="Internal Server Error")

        def signal_handler(sig, frame):
            print(f"Signal {sig} received, shutting down...")
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:  # 'RuntimeError: There is no current event loop...'
                loop = None

            # tsk = loop.create_task(self.end())
            # while not tsk.done():
            #     time.sleep(1)
            print("Shutdown complete")
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)
        # signal.signal(signal.SIGKILL, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        for tool in self.tools.values():
            logging.info(f"the tool:{tool.get_name()} has started", exc_info=True)
        uvicorn.run(self.app, host="0.0.0.0", port=self.port)



from simplegmail import Gmail
from simplegmail.message import Message
from simplegmail.label import Label
from simplegmail.attachment import Attachment
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from googleapiclient.errors import HttpError
import re
import base64
import html
import dateutil.parser as parser
from typing import Optional, List


class GmailClientWrapper(Gmail):
    def _create_message(
        self,
        sender: str,
        to: str,
        subject: str = "",
        msg_html: str = None,
        msg_plain: str = None,
        cc: List[str] = None,
        bcc: List[str] = None,
        attachments: List[str] = None,
        signature: bool = False,
        user_id: str = "me",
        reply_to: Optional[str] = None,
    ) -> dict:
        """
        Creates the raw email message to be sent.

        Args:
            sender: The email address the message is being sent from.
            to: The email address the message is being sent to.
            subject: The subject line of the email.
            msg_html: The HTML message of the email.
            msg_plain: The plain text alternate message of the email (for slow
                or old browsers).
            cc: The list of email addresses to be Cc'd.
            bcc: The list of email addresses to be Bcc'd
            attachments: A list of attachment file paths.
            signature: Whether the account signature should be added to the
                message. Will add the signature to your HTML message only, or a
                create a HTML message if none exists.
            reply_to: The id of the message that this email replies to.

        Returns:
            The message dict.

        """

        msg = MIMEMultipart("mixed" if attachments else "alternative")
        msg["To"] = to
        msg["From"] = sender
        msg["Subject"] = subject

        if cc:
            msg["Cc"] = ", ".join(cc)

        if bcc:
            msg["Bcc"] = ", ".join(bcc)

        if signature:
            m = re.match(r".+\s<(?P<addr>.+@.+\..+)>", sender)
            address = m.group("addr") if m else sender
            account_sig = self._get_alias_info(address, user_id)["signature"]

            if msg_html is None:
                msg_html = ""

            msg_html += "<br /><br />" + account_sig

        if reply_to:
            msg_id = reply_to.headers.get("Message-ID") or reply_to.id
            msg["In-Reply-To"] = msg_id
            msg["References"] = msg_id

        attach_plain = MIMEMultipart("alternative") if attachments else msg
        attach_html = MIMEMultipart("related") if attachments else msg

        if msg_plain:
            attach_plain.attach(MIMEText(msg_plain, "plain"))

        if msg_html:
            attach_html.attach(MIMEText(msg_html, "html"))

        if attachments:
            attach_plain.attach(attach_html)
            msg.attach(attach_plain)

            self._ready_message_with_attachments(msg, attachments)

        res = {"raw": base64.urlsafe_b64encode(msg.as_string().encode()).decode()}

        if reply_to:
            res["threadId"] = reply_to.thread_id

        return res

    def _build_message_from_ref(
        self,
        user_id: str,
        message_ref: dict,
        attachments: str = "reference",
        is_draft: bool = False,
    ) -> Message:
        """
        Creates a Message object from a reference.

        Args:
            user_id: The username of the account the message belongs to.
            message_ref: The message reference object returned from the Gmail
                API.
            attachments: Accepted values are 'ignore' which completely ignores
                all attachments, 'reference' which includes attachment
                information but does not download the data, and 'download' which
                downloads the attachment data to store locally. Default
                'reference'.
            is_draft: Whether the message is a draft or not.

        Returns:
            The Message object.

        Raises:
            googleapiclient.errors.HttpError: There was an error executing the
                HTTP request.

        """

        try:
            # Get message JSON
            if is_draft:
                message = (
                    self.service.users()
                    .drafts()
                    .get(userId=user_id, id=message_ref["id"])
                    .execute()["message"]
                )
            else:
                message = (
                    self.service.users()
                    .messages()
                    .get(userId=user_id, id=message_ref["id"])
                    .execute()
                )

        except HttpError as error:
            # Pass along the error
            raise error

        else:
            msg_id = message["id"]
            thread_id = message["threadId"]
            label_ids = []
            if "labelIds" in message:
                user_labels = {x.id: x for x in self.list_labels(user_id=user_id)}
                label_ids = [user_labels[x] for x in message["labelIds"]]
            snippet = html.unescape(message["snippet"])

            payload = message["payload"]
            headers = payload["headers"]

            # Get header fields (date, from, to, subject)
            date = ""
            sender = ""
            recipient = ""
            subject = ""
            msg_hdrs = {}
            cc = []
            bcc = []
            for hdr in headers:
                if hdr["name"].lower() == "date":
                    try:
                        date = str(parser.parse(hdr["value"]).astimezone())
                    except Exception:
                        date = hdr["value"]
                elif hdr["name"].lower() == "from":
                    sender = hdr["value"]
                elif hdr["name"].lower() == "to":
                    recipient = hdr["value"]
                elif hdr["name"].lower() == "subject":
                    subject = hdr["value"]
                elif hdr["name"].lower() == "cc":
                    cc = hdr["value"].split(", ")
                elif hdr["name"].lower() == "bcc":
                    bcc = hdr["value"].split(", ")

                msg_hdrs[hdr["name"]] = hdr["value"]

            parts = self._evaluate_message_payload(
                payload, user_id, message_ref["id"], attachments
            )

            plain_msg = None
            html_msg = None
            attms = []
            for part in parts:
                if part["part_type"] == "plain":
                    if plain_msg is None:
                        plain_msg = part["body"]
                    else:
                        plain_msg += "\n" + part["body"]
                elif part["part_type"] == "html":
                    if html_msg is None:
                        html_msg = part["body"]
                    else:
                        html_msg += "<br/>" + part["body"]
                elif part["part_type"] == "attachment":
                    attm = Attachment(
                        self.service,
                        user_id,
                        msg_id,
                        part["attachment_id"],
                        part["filename"],
                        part["filetype"],
                        part["data"],
                    )
                    attms.append(attm)

            return Message(
                self.service,
                self.creds,
                user_id,
                msg_id,
                thread_id,
                recipient,
                sender,
                subject,
                date,
                snippet,
                plain_msg,
                html_msg,
                label_ids,
                attms,
                msg_hdrs,
                cc,
                bcc,
            )

    def create_draft(
        self,
        sender: str,
        to: str,
        subject: str = "",
        msg_html: Optional[str] = None,
        msg_plain: Optional[str] = None,
        cc: Optional[List[str]] = None,
        bcc: Optional[List[str]] = None,
        attachments: Optional[List[str]] = None,
        signature: bool = False,
        user_id: str = "me",
        reply_to: Optional[str] = None,
    ) -> Message:
        """
        Create and insert a draft email.

        Args:
            sender: The email address the message is being sent from.
            to: The email address the message is being sent to.
            subject: The subject line of the email.
            msg_html: The HTML message of the email.
            msg_plain: The plain text alternate message of the email. This is
                often displayed on slow or old browsers, or if the HTML message
                is not provided.
            cc: The list of email addresses to be cc'd.
            bcc: The list of email addresses to be bcc'd.
            attachments: The list of attachment file names.
            signature: Whether the account signature should be added to the
                message.
            user_id: The address of the sending account. 'me' for the
                default address associated with the account.
            reply_to: The id of the message that this draft replies to.

        Returns:
            The Message object representing the draft message.

        Raises:
            googleapiclient.errors.HttpError: There was an error executing the
                HTTP request.

        """

        msg = self._create_message(
            sender,
            to,
            subject,
            msg_html,
            msg_plain,
            cc=cc,
            bcc=bcc,
            attachments=attachments,
            signature=signature,
            user_id=user_id,
            reply_to=reply_to,
        )

        try:
            req = (
                self.service.users().drafts().create(userId="me", body={"message": msg})
            )
            res = req.execute()
            return self._build_message_from_ref(
                user_id, res, "reference", is_draft=True
            )

        except HttpError as error:
            # Pass along the error
            raise error

    def send_message(
        self,
        sender: str,
        to: str,
        subject: str = "",
        msg_html: Optional[str] = None,
        msg_plain: Optional[str] = None,
        cc: Optional[List[str]] = None,
        bcc: Optional[List[str]] = None,
        attachments: Optional[List[str]] = None,
        signature: bool = False,
        user_id: str = "me",
        reply_to: Optional[str] = None,
    ) -> Message:
        """
        Sends an email.

        Args:
            sender: The email address the message is being sent from.
            to: The email address the message is being sent to.
            subject: The subject line of the email.
            msg_html: The HTML message of the email.
            msg_plain: The plain text alternate message of the email. This is
                often displayed on slow or old browsers, or if the HTML message
                is not provided.
            cc: The list of email addresses to be cc'd.
            bcc: The list of email addresses to be bcc'd.
            attachments: The list of attachment file names.
            signature: Whether the account signature should be added to the
                message.
            user_id: The address of the sending account. 'me' for the
                default address associated with the account.
            reply_to: The id of the message that this email replies to.

        Returns:
            The Message object representing the sent message.

        Raises:
            googleapiclient.errors.HttpError: There was an error executing the
                HTTP request.

        """

        msg = self._create_message(
            sender,
            to,
            subject,
            msg_html,
            msg_plain,
            cc=cc,
            bcc=bcc,
            attachments=attachments,
            signature=signature,
            user_id=user_id,
            reply_to=reply_to,
        )

        try:
            req = self.service.users().messages().send(userId="me", body=msg)
            res = req.execute()
            return self._build_message_from_ref(user_id, res, "reference")

        except HttpError as error:
            # Pass along the error
            raise error

    def get_messages(
        self,
        user_id: str = "me",
        labels: Optional[List[Label]] = None,
        query: str = "",
        max_results: int = 100,
        attachments: str = "reference",
        include_spam_trash: bool = False,
    ) -> List[Message]:
        """
        Gets messages from your account.

        Args:
            user_id: the user's email address. Default 'me', the authenticated
                user.
            labels: label IDs messages must match.
            query: a Gmail query to match.
            max_results: Maximum number of messages to return. This field defaults to 100.
            attachments: accepted values are 'ignore' which completely
                ignores all attachments, 'reference' which includes attachment
                information but does not download the data, and 'download' which
                downloads the attachment data to store locally. Default
                'reference'.
            include_spam_trash: whether to include messages from spam or trash.

        Returns:
            A list of message objects.

        Raises:
            googleapiclient.errors.HttpError: There was an error executing the
                HTTP request.

        """

        if labels is None:
            labels = []

        labels_ids = [lbl.id if isinstance(lbl, Label) else lbl for lbl in labels]

        try:
            page_token = None
            remaining = max_results
            message_refs = []

            while remaining > 0:
                retrieve_count = min(500, remaining)

                response = (
                    self.service.users()
                    .messages()
                    .list(
                        userId=user_id,
                        q=query,
                        labelIds=labels_ids,
                        maxResults=retrieve_count,
                        includeSpamTrash=include_spam_trash,
                        pageToken=page_token,
                    )
                    .execute()
                )

                if "messages" in response:  # ensure request was successful
                    message_refs.extend(response["messages"])

                page_token = response.get("nextPageToken")
                remaining -= retrieve_count

                if not page_token:
                    break

            return self._get_messages_from_refs(
                user_id,
                message_refs,
                attachments,
            )

        except HttpError as error:
            # Pass along the error
            raise error

    def add_labels(self, message: Message, labels: List[Label]):
        """
        Add labels to a message's context

        Args:
            message: The message to add labels to
            labels: A list of labels to add

        Raises:
            googleapiclient.errors.HttpError: There was an error executing the
                HTTP request.
        """
        try:
            self.service.users().threads().modify(
                userId="me",
                id=message.thread_id,
                body={"addLabelIds": [lbl.id for lbl in labels]},
            ).execute()
        except HttpError as error:
            # Pass along the error
            raise error

    def remove_labels(self, message: Message, labels: List[Label]):
        """
        Remove labels from a message's context

        Args:
            message: The message to remove labels from
            labels: A list of labels to remove

        Raises:
            googleapiclient.errors.HttpError: There was an error executing the
                HTTP request.
        """
        try:
            self.service.users().threads().modify(
                userId="me",
                id=message.thread_id,
                body={"removeLabelIds": [lbl.id for lbl in labels]},
            ).execute()
        except HttpError as error:
            # Pass along the error
            raise error

    def get_message_by_id(self, message_id: str):
        """
        Get an email message using the given id

        Args:
            message_id: message id

        Returns:
            The Message object

        Raises:
            googleapiclient.errors.HttpError: There was an error executing the
                HTTP request.
        """

        try:
            res = (
                self.service.users()
                .messages()
                .get(userId="me", id=message_id)
                .execute()
            )
            return self._build_message_from_ref(
                user_id="me", message_ref=res, attachments="reference"
            )
        except HttpError as error:
            # Pass along the error
            raise error



# pylint: disable=missing-module-docstring
import asyncio

from npiai import agent
from npiai.tools.google import GoogleCalendar
from google.oauth2.credentials import Credentials


async def main():
    credentials = Credentials.from_authorized_user_file("gc_token.json")

    async with agent.wrap(GoogleCalendar(credentials)) as gc:
        return await gc.chat(instruction="get the incoming 10 events")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import json
import os
from typing import Literal

import discord

from npiai.context import Context
from npiai.utils import logger
from npiai import FunctionTool, function
from npiai.error.auth import UnauthorizedError

__PROMPT__ = """
You are a Discord Agent helping user send/retrieve messages to/from discord channels. 

For any tool requesting a discord channel ID, you must search for it in the chat history first. 
If not found, you should ask the user to enter the ID through the `ask_human` tool.

## Example

Task: Send a DM to {{some_user}} and wait for their reply.
Workflow:
- ask_for_id({ "name": "user" })
- create_dm({ "user_id": "{{user_id}}" })
- send_message({ "channel_id": "{{dm_channel_id}}", "content": "Hi there!" })
- wait_for_reply({ "channel_id": "{{dm_channel_id}}", "message_id": "{{message_id}}" })

Task: Get the latest message from {{some_channel}}.
Workflow:
- ask_for_id({ "name": "channel" })
- fetch_history({ "channel_id": "{{channel_id}}" })

Task: Send a message to {{some_channel}}.
Workflow:
- ask_for_id({ "name": "channel" })
- send_message({ "channel_id": "{{channel_id}}", "content": "Hi there!" })
"""


class Discord(FunctionTool):
    name = "discord"
    description = "Send/Retrieve messages to/from discord channels"
    system_prompt = __PROMPT__

    _client: discord.Client
    _access_token: str

    def __init__(self, access_token: str = None):
        super().__init__()

        self._access_token = access_token or os.environ.get(
            "DISCORD_ACCESS_TOKEN", None
        )

        self._client = discord.Client(intents=discord.Intents.default())

    async def start(self):
        if self._access_token is None:
            raise UnauthorizedError("Discord credentials are not found")

        await super().start()
        await self._client.login(self._access_token)

    async def end(self):
        await super().end()
        await self._client.close()

    def parse_user(self, user: discord.User):
        return {
            "id": user.id,
            "name": user.name,
        }

    def parse_message(self, msg: discord.Message):
        return {
            "id": msg.id,
            "author": self.parse_user(msg.author),
            "content": msg.content,
            "created_at": msg.created_at.isoformat(),
            "reply_to_message_id": msg.reference.message_id if msg.reference else None,
            "mentions": [self.parse_user(user) for user in msg.mentions],
            "mention_everyone": msg.mention_everyone,
        }

    @function
    async def ask_for_id(
        self, ctx: Context, name: Literal["user", "channel", "message"]
    ):
        """
        Ask the user to provide recipient's user id, channel id, or message id.
        Args:
            ctx: NPi context.
            name: The type of id to ask for.
        """
        return await self.hitl.input(
            ctx,
            self.name,
            f"Please provide recipient's {name} ID to send messages to",
        )

    @function
    async def create_dm(self, user_id: int):
        """
        Create a direct message channel with a specific user.

        Args:
            user_id: The ID of the **user** who will receive the direct message. Note that this is not the channel ID.
        """
        user = await self._client.fetch_user(user_id)
        channel = await user.create_dm()

        return f"Direct message channel created. Channel ID: {channel.id}"

    @function
    async def fetch_history(self, channel_id: int, max_results: int = 1):
        """
        Fetch history messages from the discord channel with the given channel ID.

        Args:
            channel_id: The ID of the **channel** to send the message to. You should ask the user for it if not provided.
            max_results: The maximum number of messages to fetch.
        """
        channel = await self._client.fetch_channel(channel_id)
        messages = []

        async for msg in channel.history(limit=max_results):
            messages.append(self.parse_message(msg))

        logger.debug(
            f"[{self.name}]: Fetched {len(messages)} messages: {json.dumps(messages, indent=2, ensure_ascii=False)}"
        )

        return json.dumps(messages, ensure_ascii=False)

    @function
    async def send_message(self, channel_id: int, content: str):
        """
        Send a message to the discord channel with the given channel ID.

        Args:
            channel_id: The ID of the **channel** to send the message to. You should ask the user for it if not provided.
            content: The message to send.
        """
        channel = await self._client.fetch_channel(channel_id)
        msg = await channel.send(content)

        logger.debug(f"[{self.name}]: Sent message: (id: {msg.id}) {msg.content}")

        return f"Message sent. ID: {msg.id}"

    @function
    async def reply(self, channel_id: int, message_id: int, content: str):
        """
        Reply to a message in the discord channel with the given channel ID.

        Args:
            channel_id: The ID of the **channel** to send the message to. You should ask the user for it if not provided.
            message_id: The ID of the message to reply. You should ask the user for it if not provided.
            content: The message to reply.
        """
        channel = await self._client.fetch_channel(channel_id)
        msg = await channel.fetch_message(message_id)
        reply = await msg.reply(content)

        logger.debug(
            f"[{self.name}]: Created reply for message ID {msg.id}. Reply: (id: {reply.id}) {reply.content}"
        )

        return f"Created reply for message ID {msg.id}. Reply ID: {reply.id}"

    @function
    async def wait_for_reply(self, channel_id: int, message_id: int):
        """
        Wait for a reply to the given message in the discord channel with the given channel ID.

        Args:
            channel_id: The ID of the **channel** to send the message to. You should ask the user for it if not provided.
            message_id: The ID of the message being replied to. You should ask the user for it if not provided.
        """
        channel = await self._client.fetch_channel(channel_id)
        ref_msg = await channel.fetch_message(message_id)
        last_msg = ref_msg

        while True:
            history = channel.history(limit=10, after=last_msg.created_at)

            async for msg in history:
                # if (msg.reference and msg.reference.message_id == ref_msg.id) or \
                #     msg.mention_everyone or \
                #     any(user.id == ref_msg.author.id for user in msg.mentions):
                if msg.author.id != ref_msg.author.id and (
                    not msg.reference or msg.reference.message_id == ref_msg.id
                ):
                    return f"Received reply. {json.dumps(self.parse_message(msg))}"

                last_msg = msg

            await asyncio.sleep(3)



from abc import ABC, abstractmethod

from npiai.context import Context


class HITL(ABC):
    @abstractmethod
    async def confirm(self, ctx: Context, tool_name: str, message: str) -> bool: ...

    @abstractmethod
    async def input(self, ctx: Context, tool_name: str, message: str) -> str: ...



from .discord.app import Discord
from .github.app import GitHub
from .google.gmail.app import Gmail
from .google.calendar.app import GoogleCalendar
from .slack.app import Slack
from .twilio.app import Twilio

__all__ = [
    "Discord",
    "GitHub",
    "Gmail",
    "GoogleCalendar",
    "Slack",
    "Twilio",
]



import pathlib
import tempfile
import os
import uuid
from urllib.request import urlretrieve
import traceback
from playwright.async_api import (
    async_playwright,
    Playwright,
    Browser,
    BrowserContext,
    Page,
    FileChooser,
)

__BROWSER_UTILS_VERSION__ = "0.0.2"


def _prepare_browser_utils():
    # path to the js bundle
    cache_dir = pathlib.Path(tempfile.gettempdir()) / ".npi"
    js_path = cache_dir / f"chrome-utils@{__BROWSER_UTILS_VERSION__}.js"

    if js_path.exists():
        return js_path

    os.makedirs(cache_dir, exist_ok=True)

    urlretrieve(
        f"https://unpkg.com/@npi-ai/browser-utils@{__BROWSER_UTILS_VERSION__}/dist/index.global.js",
        js_path,
    )

    return js_path


class PlaywrightContext:
    def __init__(
        self,
        headless: bool = True,
    ):
        """
        Initialize a Playwright context

        Args:
            headless: Whether to run playwright in headless mode
        """
        self.headless = headless
        self.ready = False
        self.playwright: Playwright | None = None
        self.browser: Browser | None = None
        self.context: BrowserContext | None = None
        self.page: Page | None = None

    async def start(self):
        """Start the Playwright chrome"""
        self.playwright = await async_playwright().start()
        self.browser = await self.playwright.chromium.launch(
            headless=self.headless,
            args=["--disable-gpu", "--single-process"],
        )

        self.context = await self.browser.new_context(
            locale="en-US",
            bypass_csp=True,
            **self.playwright.devices["Desktop Edge"],
        )
        # self.context.set_default_timeout(3000)
        await self.context.add_init_script(path=_prepare_browser_utils())
        await self.context.add_init_script(
            script="""window.npi = new window.BrowserUtils()"""
        )

        self.page = await self.context.new_page()
        self.page.on("filechooser", self.on_filechooser)
        self.page.on("popup", self.on_popup)

        self.ready = True

    async def on_filechooser(self, chooser: FileChooser):
        """
        Callback function invoked when an input:file is clicked

        Args:
            chooser: FileChooser instance
        """
        await chooser.set_files("")

    async def on_popup(self, popup: Page):
        """
        Callback function invoked when a tab is opened

        Args:
            popup: Page instance
        """
        print(f"popup {popup}")
        self.page = popup

    async def stop(self):
        """
        Dispose the chrome tools
        """
        await self.playwright.stop()



from typing import Dict, Any, Type

from pydantic import BaseModel


def sanitize_schema(model: Type[BaseModel]) -> Dict[str, Any]:
    schema = model.model_json_schema()

    # remove unnecessary title
    schema.pop("title", None)

    for prop in schema.get("properties", {}).values():
        prop.pop("title", None)

        # use a more compact format for optional fields
        if (
            "anyOf" in prop
            and len(prop["anyOf"]) == 2
            and prop["anyOf"][1]["type"] == "null"
        ):
            # copy the first type definition to props
            t = prop["anyOf"][0]
            for k, v in t.items():
                prop[k] = v

            prop.pop("anyOf", None)

            if prop.get("default") is None:
                prop.pop("default", None)

    # remove empty properties
    if len(schema.get("properties", {})) == 0:
        schema.pop("properties", None)

    return schema



from typing import Dict

from npiai.context import Context
from npiai.cloud import Client


class CloudContext(Context):

    def __init__(self, req, client: Client | None = None):
        if not client:
            client = Client(access_token=req.headers.get("x-npi-access-token"))
        super().__init__()
        self.client = client

    def credentials(self, app_code: str) -> Dict[str, str]:
        return self.client.get_credentials(app_code)



from enum import Enum
from typing import List, Union
import os
import asyncio
from litellm import acompletion, ModelResponse, CustomStreamWrapper


class Provider(Enum):
    OpenAI = 1
    Llama = 2
    Anthropic = 3
    Groq = 4
    AzureOpenAI = 5


class LLM:
    def __init__(self, api_key: str, model: str, provider: Provider):
        self.model = model
        self.api_key = api_key
        self.provider = provider

    def default_model(self) -> str:
        return self.model

    def get_provider(self) -> Provider:
        return self.provider

    # TODO: kwargs typings
    async def completion(self, **kwargs) -> Union[ModelResponse, CustomStreamWrapper]:
        return await acompletion(model=self.model, api_key=self.api_key, **kwargs)

    def completion_sync(self, **kwargs) -> Union[ModelResponse, CustomStreamWrapper]:
        return asyncio.run(self.completion(**kwargs))


class OpenAI(LLM):
    def __init__(self, api_key: str, model: str):
        super().__init__(api_key=api_key, model=model, provider=Provider.OpenAI)


class Anthropic(LLM):
    def __init__(self, api_key: str, model: str):
        super().__init__(api_key=api_key, model=model, provider=Provider.Anthropic)


class AzureOpenAI(LLM):
    def __init__(
        self, api_key: str, api_base: str, api_version: str, deployment_name: str
    ):
        os.environ["AZURE_API_BASE"] = api_base
        os.environ["AZURE_API_VERSION"] = api_version
        super().__init__(f"azure/{deployment_name}", api_key, Provider.AzureOpenAI)



from npiai.core.browser import PlaywrightContext, NavigatorAgent
from .hitl import HITL
from .base import BaseTool

__all__ = [
    "BaseTool",
    "PlaywrightContext",
    "NavigatorAgent",
    "HITL",
]



from typing import Type
import json
import signal
import sys
import re

import uvicorn
import logging
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse

from npiai.core.tool._function import BaseTool
from npiai.context import Context


class Container:
    def __init__(self, tool_cls: Type[BaseTool]):
        self.tool_cls = tool_cls

    def schemas(self):
        obj = self.tool_cls()
        print(obj.schema())

    def run(self, port: int = 19140):
        app = FastAPI()

        @app.api_route(
            "/{full_path:path}", methods=["GET", "POST", "PUT", "DELETE", "PATCH"]
        )
        async def root(full_path: str, request: Request):
            # ctx = Context.from_request(req=request)
            ctx = Context()
            tool = ctx.get_tool()
            if tool is None:
                tool = self.tool_cls.from_context(ctx=ctx)
                await tool.start()

            try:
                ctx.entry()
                method = self.convert_camel_to_snake(full_path)
                match request.method:
                    case "POST":
                        args = await request.json()
                    case "GET":
                        args = {k: v for k, v in request.query_params.items()}
                    case _:
                        return JSONResponse(
                            {"error": "Method not allowed"}, status_code=405
                        )
                res = await tool.exec(ctx, method, args)
                return JSONResponse(
                    content=json.loads(res),
                    headers={
                        "X-Context-Id": ctx.id,
                    },
                )
            except Exception as e:
                logging.error(f"Failed to process request: {e}", exc_info=True)
                return JSONResponse({"error": f"internal error: {e}"}, status_code=500)
            finally:
                ctx.exit()

        def signal_handler(sig, frame):
            print(f"Signal {sig} received, shutting down...")
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)
        # signal.signal(signal.SIGKILL, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

        uvicorn.run(app, host="0.0.0.0", port=port)

    @staticmethod
    def convert_camel_to_snake(name: str) -> str:
        return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()


if __name__ == "__main__":
    from npiai.tools import GitHub

    container = Container(GitHub)
    container.run()



from typing import Callable, Awaitable
import functools
import asyncio


def to_async_fn(fn: Callable) -> Callable[..., Awaitable]:
    @functools.wraps(fn)
    async def wrapper(*args, **kwargs):
        res = fn(*args, **kwargs)
        if asyncio.iscoroutine(res):
            return await res
        return res

    return wrapper



import asyncio
import uuid

from playground.proto import playground_pb2


class ActionResult:

    def __init__(self, result: playground_pb2.ActionResultRequest):
        self.result = result

    def is_approved(self):
        return self.result.action_result != "denied"

    def has_message(self):
        pass


class Callable:

    def __init__(
        self, msg: str = None, action: playground_pb2.ActionRequiredResponse = None
    ):
        self.msg = msg
        self.called = False
        self.action = action
        if action:
            self.__type = playground_pb2.ResponseCode.ACTION_REQUIRED
        else:
            self.__type = playground_pb2.ResponseCode.MESSAGE
        self.__id = str(uuid.uuid4())
        self.__future = asyncio.get_event_loop().create_future()

    def id(self):
        return self.__id

    def client_response(self):
        if self.__type == playground_pb2.ResponseCode.MESSAGE:
            return playground_pb2.ChatResponse(message=self.msg)
        elif self.__type == playground_pb2.ResponseCode.ACTION_REQUIRED:
            return self.action
        return self.__type

    def type(self) -> playground_pb2.ResponseCode:
        return self.__type

    def message(self) -> str:
        return self.msg

    def callback(self, result: playground_pb2.ActionResultRequest):
        self.__future.set_result(result)

    async def wait(self) -> ActionResult:
        req = await self.__future
        return ActionResult(req)



import asyncio

from npiai import agent
from npiai.tools import GitHub


async def main():
    async with agent.wrap(GitHub()) as github:
        return await github.chat(
            'Find the issue in the repo idiotWu/npi-test titled "Test Issue for NPi" and change the body to "Hello World"'
        )


if __name__ == "__main__":
    asyncio.run(main())



import json
import os
import re
import tempfile
import pathlib
from typing import Literal

from playwright.async_api import TimeoutError, Error, Locator
from markdownify import MarkdownConverter
from slugify import slugify

from npiai import function, BrowserTool
from npiai.context import Context
from npiai.core.browser import NavigatorAgent
from npiai.utils import logger
from npiai.error.auth import UnauthorizedError

__SYSTEM_PROMPT__ = """
You are a Twitter Agent helping user retrieve and manage tweets. 

For any given task, you should first check if you are on the correct page. If not, you should use the navigator to go to the target page. 

For any action regarding posting a tweet or deleting a tweet, you should ask for user confirmation before proceeding. Otherwise, you should not stop until the task is complete.

For tasks requiring retrieval of the page contents, you may call the `get_text()` tool if no other suitable tool is available.

Note that the navigator is designed to finish arbitrary tasks by simulating user inputs, you may use it as the last resort to fulfill the task if all other tools are not able to complete the task. In this case, you may also give the navigator a rough plan as the hint.

However, the navigator is unable to retrieve the content of the tweets. If the task includes the analysis of a specific tweet, you should break it into several subtasks and coordinate tools to finish it. For example:

Task: reply to the latest tweet of @Dolphin_Wood

Steps:
1. goto_profile({ "username": "Dolphin_Wood" })
2. get_tweets({ "max_results": 1 })
3. open_tweet({ "url": "{{tweet_url}}" })
4. navigator({ "task": "Reply to current tweet with {{reply_content}}. Ask for user confirmation before posting." })

Task: get the last 10 tweets by @Dolphin_Wood

Steps:
1. goto_profile({ "username": "Dolphin_Wood" })
2. get_tweets({ "max_results": 10 })
3. Repeat the following steps until 10 tweets are found. Remember that you should exclude the pinned tweet.
  3.1. load_more_tweets()
  3.2. get_tweets({ "max_results": 10 })
"""

__ROUTES__ = {"login": "https://x.com/", "home": "https://x.com/home"}

from npiai.llm import LLM


class ImageFilterConverter(MarkdownConverter):
    def process_text(self, el):
        text = el.text

        if text == "Quote":
            return "\nQuote Tweet: "

        # remove messages inside the video component
        if el.find_parent(
            name="div", attrs={"data-testid": re.compile(r"videoComponent|tweetPhoto")}
        ):
            return ""

        parent = el.find_parent(
            name="div", attrs={"data-testid": re.compile(r"reply|retweet|like")}
        )

        if not parent:
            return super().process_text(el)

        count = text or "0"
        # annotate footer icons
        match parent.attrs.get("data-testid"):
            case "reply":
                return f" Replies: {count} "
            case "retweet":
                return f" Retweets: {count} "
            case "like":
                return f" Likes: {count} "
            case _:
                return super().process_text(el)

    def convert_a(self, el, text, convert_as_inline):
        if re.search(r"analytics", el.attrs.get("href", "")):
            return f' Views: {el.text or "0"}'
        return super().convert_a(el, text, convert_as_inline)

    def convert_img(self, el, text, convert_as_inline):
        src = el.attrs.get("src", "")
        # remove profile/card pictures
        if re.search(r"profile_images|card_img", src):
            return ""
        # remove emoji images
        if re.search(r"emoji", src):
            return el.attrs.get("alt")

        return super().convert_img(el, text, convert_as_inline)


def html_to_md(html: str, **options) -> str:
    return ImageFilterConverter(**options).convert(html)


class Twitter(BrowserTool):
    name = "twitter"
    description = "retrieve and manage tweets"
    system_prompt = __SYSTEM_PROMPT__

    _username: str
    _password: str

    def __init__(
        self,
        username: str = None,
        password: str = None,
        navigator_llm: LLM = None,
        headless: bool = True,
    ):
        super().__init__(
            headless=headless,
        )

        self._username = username or os.environ.get("TWITTER_USERNAME", None)
        self._password = password or os.environ.get("TWITTER_PASSWORD", None)
        self.ctx = None

        self.add_tool(NavigatorAgent(llm=navigator_llm, playwright=self.playwright))

    async def start(self):
        if self._username is None:
            raise UnauthorizedError("No Twitter username provided.")

        if self._password is None:
            raise UnauthorizedError("No Twitter password provided.")

        if not self._started:
            await super().start()
            await self._login(self.ctx)

    async def _login(self, ctx: Context | None = None):
        state_file = (
            pathlib.Path(tempfile.gettempdir())
            / f"{slugify(self._username)}/twitter_state.json"
        )
        if os.path.exists(state_file):
            with open(state_file, "r") as f:
                state = json.load(f)
                await self.playwright.context.add_cookies(state["cookies"])
            await self.playwright.page.goto(__ROUTES__["home"])
            try:
                # validate cookies
                await self.playwright.page.wait_for_url(__ROUTES__["home"])
                logger.debug("Twitter cookies restored.")
                return
            except TimeoutError:
                # cookies expired, continue login process
                logger.debug("Twitter cookies expired. Continue login process.")

        await self.playwright.page.goto(__ROUTES__["login"])
        await self.playwright.page.get_by_test_id("loginButton").click()
        await self.playwright.page.get_by_label("Phone, email, or username").fill(
            self._username
        )
        await self.playwright.page.get_by_role("button", name="Next").click()

        # check if additional credentials (i.e, username) is required
        await self._check_additional_credentials(ctx)

        await self.playwright.page.get_by_label("Password", exact=True).fill(
            self._password
        )
        await self.playwright.page.get_by_test_id("LoginForm_Login_Button").click()

        # check again if additional credentials (i.e, phone number) is required
        await self._check_additional_credentials(ctx)

        # now we should be directed to twitter home
        await self.playwright.page.wait_for_url(__ROUTES__["home"])

        # save state
        save_dir = os.path.dirname(state_file)
        os.makedirs(save_dir, exist_ok=True)
        await self.playwright.context.storage_state(path=state_file)

    async def _get_additional_cred_input(self) -> Locator | None:
        await self.playwright.page.wait_for_timeout(1000)
        cred_input = self.playwright.page.get_by_test_id("ocfEnterTextTextInput")

        try:
            if await cred_input.count() != 0:
                return cred_input
        except Error:
            return None

    async def _check_additional_credentials(self, ctx: Context | None = None):
        cred_input = await self._get_additional_cred_input()

        if cred_input:
            label = self.playwright.page.locator(
                'label:has(input[data-testid="ocfEnterTextTextInput"])'
            )
            cred_name = await label.text_content()
            cred = await self._request_additional_credentials(cred_name, ctx)
            await cred_input.fill(cred)
            await self.playwright.page.get_by_test_id("ocfEnterTextNextButton").click()

            if await self._get_additional_cred_input() is not None:
                raise UnauthorizedError(
                    "Unable to login to Twitter. Please try again with the correct credentials."
                )

    async def _request_additional_credentials(
        self, cred_name: str, ctx: Context | None = None
    ) -> str:
        if ctx is None:
            raise UnauthorizedError(
                f"Unable to login to Twitter. Please replace username with {cred_name} and try again."
            )

        return await self.hitl.input(
            ctx,
            self.name,
            f"Please enter {cred_name} to continue the login process.",
        )

    @function
    async def search(
        self, query: str, sort_by: Literal["hottest", "latest"] = "hottest"
    ):
        """
        Search Twitter for tweets or users matching the given query.

        Args:
            query: Twitter search query.
            sort_by: Sort results by this parameter. Default is "hottest".
        """
        await self.playwright.page.goto("https://x.com/explore")
        search_input = self.playwright.page.get_by_test_id("SearchBox_Search_Input")
        await search_input.fill(query)
        await search_input.press("Enter")

        if sort_by == "latest":
            await self.playwright.page.wait_for_timeout(1000)
            await self.playwright.page.get_by_role("tab", name="Latest").click()

        return f"The search result page is opened. Please use other tools to retrieve the results. Current page: {await self.get_page_title()}"

    @function
    async def goto_profile(self, username: str):
        """
        Goto Twitter user's profile page.

        Args:
            username: Twitter username.
        """
        username = username.replace("@", "")
        await self.playwright.page.goto(f"https://x.com/{username}")

        return f"The profile page for user {username} is opened. Current page: {await self.get_page_title()}"

    @function
    async def goto_notifications(self):
        """Goto Twitter user's notifications page."""
        await self.playwright.page.goto("https://x.com/notifications/mentions")

        return f"The notifications page is opened. Current page: {await self.get_page_title()}"

    @function
    async def reply(self, url: str, content: str):
        """
        Open the given tweet and reply to it.

        Args:
            url: URL of the tweet you want to reply to.
            content: Reply content.
        """
        await self.playwright.page.goto(url)
        reply_input = self.playwright.page.get_by_test_id("tweetTextarea_0")
        await reply_input.click()
        await self.playwright.page.wait_for_timeout(300)
        await reply_input.fill(content)
        await self.playwright.page.get_by_test_id("tweetButtonInline").click()

        return f"Replied to tweet {url} with content: {content}"

    @function
    async def get_current_page(self):
        """Get the title and url of the current page."""
        result = {
            "title": await self.playwright.page.title(),
            "url": self.playwright.page.url,
        }
        return json.dumps(result, ensure_ascii=False)

    @function
    async def get_tweets(self, max_results: int = -1) -> str:
        """
        Retrieve tweets on the current page.
        You should ensure you are on the correct page (home, user profile, etc.) before calling this tool.

        Args:
            max_results: Maximum number of tweets to return. Pass -1 for no limit.
        """
        try:
            # match unvisited tweets only
            tweets = self.playwright.page.locator(
                '[data-testid="tweet"]:not([data-visited])'
            )
            await tweets.first.wait_for(state="attached", timeout=180_000)

            logger.debug(f"{await tweets.count()} tweets found.")
        except TimeoutError:
            return "No tweets found."

        results = []

        for tweet in await tweets.all():
            try:
                # skip ads
                if await tweet.get_by_text("Ad", exact=True).count() > 0:
                    logger.debug(f"Skipping ad: {await tweet.text_content()}")
                    continue
                # skip spaces
                if (
                    await tweet.get_by_role(
                        "button", name="Play recording", exact=True
                    ).count()
                    > 0
                ):
                    logger.debug(f"Skipping space: {await tweet.text_content()}")
                    continue
                # author = await tweet.get_by_test_id('User-Name').first.text_content()
                # content = author + ': ' + await tweet.get_by_test_id('tweetText').first.text_content()
                # # extract social context
                # social_context = tweet.get_by_test_id('socialContext')
                # if await social_context.count() == 1:
                #     content = await social_context.text_content() + '\n' + content
                # # extract media
                # media = tweet.locator(
                #     '[data-testid="videoPlayer"]:has([src]), \
                #      [data-testid="tweetPhoto"]:not(:has([data-testid="videoPlayer"])):has([src])'
                # )
                # if await media.count() > 0:
                #     content += ' ' + await media.evaluate_all(
                #         """elems => {
                #             return elems.map(el => {
                #                 const mediaType = el.getAttribute('data-testid') === 'tweetPhoto' ? 'image' : 'video';
                #                 const src = el.querySelector('[src]')?.src;
                #                 return `![${mediaType}][${src}]`;
                #             }).join(' ')
                #         }"""
                #     )
                # # extract retweet
                # retweet = tweet.locator('div[role="link"]:has([data-testid="tweetText"])')
                # if await retweet.count() == 1:
                #     content += '\n Retweet: ' + await retweet.text_content()

                # extract link
                link = await tweet.locator('a[href*="/status/"]').first.get_attribute(
                    "href"
                )
                results.append(
                    {
                        "link": f"https://x.com{link}",
                        "content": html_to_md(await tweet.inner_html()),
                    }
                )
            except TimeoutError as e:
                logger.error(e)
            finally:
                if -1 < max_results <= len(results):
                    break

        # mark as visited
        await tweets.evaluate_all(
            'elems => elems.forEach(el => el.setAttribute("data-visited", "true"))'
        )

        logger.debug(
            f"{len(results)} tweets retrieved: {json.dumps(results, indent=2, ensure_ascii=False)}"
        )

        return json.dumps(results, ensure_ascii=False)

    @function
    async def load_more_tweets(self) -> str:
        """Scroll to bottom of the page to load more tweets."""
        if not await self.is_scrollable():
            return "All tweets loaded. No more tweets."

        await self.playwright.page.evaluate(
            """() => {
                const tweets = document.querySelectorAll('[data-testid="tweet"]');
                tweets[tweets.length - 1]?.scrollIntoView();
            }"""
        )

        await self.playwright.page.wait_for_timeout(3000)

        return "More tweets loaded."

    @function
    async def open_tweet(self, url: str) -> str:
        """
        Open a tweet with the given URL.

        Args:
            url: URL of the tweet you want to open.
        """
        await self.playwright.page.goto(url, timeout=10_000)

        return f"Tweet opened. Current URL: {await self.get_page_url()}, page title: {await self.get_page_title()}"


__all__ = ["Twitter"]



import asyncio

from npiai import agent
from npiai.tools.google import Gmail
from google.oauth2.credentials import Credentials


async def main():
    credentials = Credentials.from_authorized_user_file("gmail_token.json")

    async with agent.wrap(Gmail(credentials)) as gmail:
        return await gmail.chat(
            "Send an email to daofeng@npi.ai inviting him to join an AI meetup and wait for his reply. The date candidates are: Monday, Tuesday, Wednesday"
        )


if __name__ == "__main__":
    asyncio.run(main())



import questionary

from termcolor import colored
from npiai.core import HITL
from npiai.context import Context


class ConsoleHandler(HITL):
    async def confirm(self, ctx: Context, tool_name: str, message: str) -> bool:
        # print(colored(f'[{tool_name}]: Confirmation required', color='green', attrs=['bold']))
        print(colored(f"[{tool_name}]: {message}", color="magenta", attrs=["bold"]))
        return await questionary.confirm("Continue?").ask_async()

    async def input(self, ctx: Context, tool_name: str, message: str) -> str:
        # print(colored(f'[{tool_name}]: Additional information required', color='green', attrs=['bold']))
        print(colored(f"[{tool_name}]: {message}", color="magenta", attrs=["bold"]))
        return await questionary.text("Type your response:").ask_async()



# TODO: recreate navigator using function callings

import json
import re
from textwrap import dedent
from typing import Union, List, Literal, Tuple

from playwright.async_api import Error

from typing_extensions import NotRequired, TypedDict

from npiai.llm import LLM
from npiai.utils import logger
from npiai.context import Context, Task
from npiai.core.browser import PlaywrightContext
from npiai.core.tool._browser import BrowserTool
from npiai.core.tool._agent import BrowserAgentTool

__PROMPT__ = """
Imagine that you are imitating humans doing web navigation for a task step by step. I will provide you with the following context and you should call the best tool to fulfill the task.

## Provided Context

- A screenshot of the target page in the current viewport.
- An annotated screenshot where the interactive elements are surrounded with rectangular bounding boxes in different colors. At the top left of each bounding box is a small rectangle in the same color as the bounding box. This is the label and it contains a number indicating the ID of that box. The label number starts from 0.
- The title of the target page.
- The task I want to perform.
- Up to 10 previous actions (you should use them to avoid duplicates). Especially, each action in the history will have an "element" field specifying the target element of the action.
- An array of the interactive elements inside the current viewport.
- An array of newly added elements' ID since the last action. These elements reflect the impact of your last action.

## Element Object

The original HTML elements are described as the following JSON objects:

type Element = {
  id: string; // The ID of the element
  tag: string; // The tag of the element
  role: string | null; // The WAI-ARIA accessible role of the element
  accessibleName: string; // The WAI-ARIA accessible name of the element
  accessibleDescription: string; // The WAI-ARIA accessible description of the element
  attributes: Record<string, string>; // Some helpful attributes of the element
  options?: string[]; // Available options of an <select> element. This property is only provided when the element is a <select> element.
}

## Instructions

Below is the instructions for you to outline the next actions of current stage:
- Examine the screenshots and the provided HTML, and then think about what the current page is.
- Validate previous actions using the screenshots. You should try to re-run the actions that are not successfully performed.
- Pay more attention to elements that overlap on top of other elements (i.e., is visible in the screenshots) and consider performing actions on these visible elements first. Remember that you are only given the elements within the current viewport.
- Go through the `role`, `accessibleName`, and `accessibleDescription` properties to grab semantic information of the elements.
- Pay attention to the newly added elements (e.g., dropdown list) since they may related to the next action.
- Analyze the previous actions and their intention. Particularly, focus on the last action, which may be more related to what you should do now as the next step.
- Think if you need to close the dialogs before interacting with the elements below it. This is necessary when the dialog overlaps with other elements.
- Think if it is necessary to scroll the page down to see more contents. Scrolling could be useful when you find that the required element is beyond current viewport.
- When working on an input of a combobox or a datepicker, it may be useful to perform a click action first.
- If the target element is overlapped by some dialogs, it may be useful to close these dialogs first.
- Briefly describe the current screenshots and the provided HTML in the "observation" field. If there are previous actions, you need to briefly describe the previous actions and the page changes.
- Based on the observation, think about what you need to do next to complete the instructions, and output your thoughts in the "thoughts" field.
- Decide the best action for the next step.

## Conventions

Here are some conventions to choose a suitable tool:
- If you need to click on an element, you should set the `action.type` field to "click";
- If you need to press the enter key on an input or textarea to trigger a search or submit a form, you should set the `action.type` to "enter".
- If you need to type or fill something into an input box, you should set the `action.type` to "fill" and specify the text in the `value` field.
- If you need to select a value for a <select> element, you should set the `action.type` to "select" and specify the value in the `value` field.
- If the page is scrollable and the task requires actions beyond the provided viewport, you can set the `action` to "scroll" to scroll down to reveal the rest of the page.
- If you think you have missed something and need to scroll back to the top of the page to start over, you can set the `action` to "back_to_top".
- If you are performing a critical action, such as submitting a form to place an order, clicking a "Send" button to send a message, or clicking a "Save" button to save the form, you should set the `action.type` to "confirmation" and ask for user confirmation.
- If you need more information from the user to fulfill the task, such as requesting the user's name and making a choice, you should set the `action.type` to "human-intervention" and wait for user response.
- If the whole task is done, you should set the `action.type` to "done" not call any further tools.

## Response Format

You must exactly follow the following response format and avoid adding any explanation or additional text in your response. Particularly, do not put any comments in the response.

{
  "observation": string,
  "thoughts": string,
  "action": {
      "description": string,
      "type": 'fill' | 'click' | 'enter' | 'select' | 'scroll' | 'back-to-top' | 'confirmation' | 'human-intervention' | 'done',
      "value"?: string, // input value for 'type' or 'select' actions, if applicable
      "id"?: string, // element ID, unnecessary for the 'scroll' action
    }
}

## Input Example

Task: Search for the answer to everything
Scrollable: yes
Previous Actions: [{...}]
Elements: [{...}]
Newly Added Elements' ID: ["0", "1", "2", ...]

## Response Example

{
  "observation": "This is the Google search page with a search box in the center",
  "thoughts": "I should type the query text in the search box",
  "action": {
    "description": "Type 'the answer to everything' into the search box",
    "type": "fill",
    "value": "the answer to everything",
    "id": "42"
  }
}
"""


class Action(TypedDict):
    description: str
    type: Literal[
        "type",
        "click",
        "enter",
        "select",
        "scroll",
        "back-to-top",
        "confirmation",
        "human-intervention",
        "not-found",
        "done",
    ]
    value: NotRequired[str]
    id: NotRequired[str]
    element: NotRequired[dict]  # element json used for recording


class Response(TypedDict):
    observation: str
    thoughts: str
    action: Action


def _parse_response(response: str) -> Union[Response, None]:
    try:
        match = re.match(r"```.*\n([\s\S]+)```", response)
        data = json.loads(match.group(1)) if match else json.loads(response)
        return data
    except json.JSONDecodeError:
        return None


class NavigatorAgent(BrowserAgentTool):
    name: str = "navigator"

    def __init__(
        self,
        llm: LLM,
        playwright: PlaywrightContext,
        max_steps: int = 42,
    ):
        self._browser_app = BrowserTool(
            name="navigator",
            description="Perform any task by simulating keyboard/mouse interaction on a specific web page. "
            "If the some action needs user confirmation, please specify them.",
            system_prompt=__PROMPT__,
            playwright=playwright,
        )

        super().__init__(
            tool=self._browser_app,
            llm=llm,
        )

        self.max_steps = max_steps

    # navigator uses shared playwright context, so we don't need to start it again here
    async def start(self):
        pass

    async def end(self):
        pass

    async def generate_user_prompt(self, task: str, history: List[Response]):
        await self._browser_app.clear_bboxes()
        raw_screenshot = await self._browser_app.get_screenshot()
        elements, added_ids = await self._browser_app.get_interactive_elements(
            raw_screenshot
        )

        user_prompt: str = dedent(
            f"""
            Page Title: {await self._browser_app.get_page_title()}
            Task: {task}
            Scrollable: {await self._browser_app.is_scrollable()}
            Previous Actions: {json.dumps(history[-10:])}
            Elements: {json.dumps(elements)}
            Newly Added Elements' ID: {json.dumps(added_ids)}
            """
        )

        # print(user_prompt)

        annotated_screenshot = await self._browser_app.get_screenshot()

        if not raw_screenshot or not annotated_screenshot:
            return {
                "role": "user",
                "content": user_prompt,
            }

        return {
            "role": "user",
            "content": [
                {
                    "type": "text",
                    "text": user_prompt,
                },
                {
                    "type": "image_url",
                    "image_url": {
                        "url": raw_screenshot,
                    },
                },
                {
                    "type": "image_url",
                    "image_url": {
                        "url": annotated_screenshot,
                    },
                },
            ],
        }

    async def chat(self, ctx: Context, instruction: str) -> str:
        history: List[Response] = []

        step = 0

        while True:
            msg = ctx.fork(instruction)
            msg.append(
                {
                    "role": "system",
                    "content": self._browser_app.system_prompt,
                }
            )
            msg.append(await self.generate_user_prompt(instruction, history))

            response_str = await self._call_llm(ctx, msg)
            response = _parse_response(response_str)

            if not response:
                # try again if the response can't be parsed correctly
                continue

            result, elem_json = await self._run_action(response["action"], ctx)
            logger.info(result)

            if not result:
                # requires further intervention if the action is not executable
                return response_str

            # remove element id from history to reduce noise
            response["action"].pop("id", None)
            response["action"]["element"] = elem_json
            history.append(response)
            step += 1

            if step > self.max_steps:
                return f"Maximum number of steps reached. Last response was: {response_str}"

    async def _call_llm(self, ctx: Context, message: Task) -> str:
        """
        Call llm for one round with the given prompts

        Args:
            message: ThreadMessage context

        Returns:
            response message
        """
        response = await self.llm.completion(
            messages=message.conversations(),
            max_tokens=4096,
        )

        response_message = response.choices[0].message

        message.step(response_message)

        if not response_message.content:
            raise Exception(f"{self.name}: No response message")

        logger.debug(response_message.content + "\n")

        return response_message.content

    async def _run_action(
        self, action: Action, ctx: Context
    ) -> Tuple[Union[str, None], dict]:
        """
        Run the given action

        Args:
            action: Action to run

        Returns:
            [0]: response message if the action is executable, None otherwise
            [1]: element json
        """
        await self._browser_app.clear_bboxes()
        await self._browser_app.init_observer()

        elem = (
            await self._browser_app.get_element_by_marker_id(action["id"])
            if "id" in action
            else None
        )
        elem_json = await self._browser_app.element_to_json(elem) if elem else None

        call_msg = f'[{self.name}]: {action["type"]} - {action["description"]}'

        logger.info(call_msg)
        await ctx.send(call_msg)

        match action["type"]:
            case "click":
                result = await self._browser_app.click(elem)
            case "enter":
                result = await self._browser_app.enter(elem)
            case "fill":
                result = await self._browser_app.fill(elem, action["value"])
            case "select":
                result = await self._browser_app.select(elem, action["value"])
            case "scroll":
                result = await self._browser_app.scroll()
            case "back-to-top":
                result = await self._browser_app.back_to_top()
            case "confirmation":
                result = await self.hitl.confirm(ctx, self.name, action["description"])
            case "human-intervention":
                result = await self.hitl.input(ctx, self.name, action["description"])
            case "done":
                result = None
            case _:
                raise Exception(f"{self.name}: Unknown action: {action}")

        if elem:
            await elem.dispose()

        try:
            await self._browser_app.wait_for_stable()
        except Error:
            # FIXME: if the action triggers a navigator, we will receive an error showing "Page.evaluate:
            #  Execution context was destroyed, most likely because of a navigation"
            await self._browser_app.playwright.page.wait_for_timeout(3000)

        return result, elem_json



from npiai.core import NavigatorAgent
from npiai import LLM, function, BrowserTool

__SYSTEM_PROMPT__ = """
You are a general chromium-based autonomous agent helping user to finish any task on any webpage. For a given task,
 you should first go to the appropriate web page, and then pass the task to the navigator to fulfill it.
"""


class Chromium(BrowserTool):
    name = "chrome"
    description = "Perform any task on any webpage"
    system_prompt = __SYSTEM_PROMPT__

    def __init__(self, navigator_llm: LLM = None, headless: bool = True):
        super().__init__(
            headless=headless,
        )
        self.add_tool(
            NavigatorAgent(
                llm=navigator_llm,
                playwright=self.playwright,
            )
        )

    @function
    async def goto(self, url: str):
        """
        Open the given URL in chromium.

        Args:
            url: The URL to open.
        """
        await self.playwright.page.goto(url)

        return f"Opened {await self.get_page_url()}, page title: {await self.get_page_title()}"



import asyncio

from npiai import agent
from npiai.tools import GitHub


async def main():
    async with agent.wrap(GitHub()) as github:
        return await github.chat(
            'Create a pull request in idiotWu/npi-test from "npi-test" branch to "main" branch with random title and body'
        )


if __name__ == "__main__":
    asyncio.run(main())



import os
from typing import List

from pydantic import create_model, Field

from npiai.llm import LLM, OpenAI
from npiai.types import FunctionRegistration
from npiai.context import Context, Task
from npiai.core.base import BaseAgentTool
from npiai.core.hitl import HITL
from npiai.core.tool._function import FunctionTool
from npiai.core.tool._browser import BrowserTool
from npiai.utils import sanitize_schema

from litellm.types.completion import (
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
)


class AgentTool(BaseAgentTool):
    llm: LLM
    _tool: FunctionTool

    def __init__(self, tool: FunctionTool, llm: LLM = None):
        super().__init__()
        self.name = f"{tool.name}_agent"
        self._tool = tool
        self.llm = llm or OpenAI(
            api_key=os.environ.get("OPENAI_API_KEY", None), model="gpt-4o"
        )
        self.description = tool.description
        self.provider = tool.provider

    def unpack_functions(self) -> List[FunctionRegistration]:
        # Wrap the chat function of this agent to FunctionRegistration
        model = create_model(
            f"{self.name}_agent_model",
            instruction=(
                str,
                Field(
                    description=f"The task you want {self._tool.name} to do or "
                    f"the message you want to chat with {self._tool.name}"
                ),
            ),
        )

        fn_reg = FunctionRegistration(
            fn=self.chat,
            name="chat",
            ctx_variables=[],
            ctx_param_name="ctx",
            description=f"This is an api of an AI Assistant, named {self.name}, the abilities of the assistant is:\n "
            f"{self.description}\n"
            f"You can use this function to direct the assistant to accomplish task for you.",
            model=model,
            schema=sanitize_schema(model),
        )

        return [fn_reg]

    def use_hitl(self, hitl: HITL):
        super().use_hitl(hitl)
        self._tool.use_hitl(hitl)

    async def start(self):
        await self._tool.start()

    async def end(self):
        await self._tool.end()

    async def chat(
        self,
        ctx: Context,
        instruction: str,
    ) -> str:
        task = Task(goal=instruction)
        ctx = ctx.fork(task)
        if self._tool.system_prompt:
            await task.step(
                [
                    ChatCompletionSystemMessageParam(
                        role="system", content=self._tool.system_prompt
                    )
                ]
            )

        await task.step(
            [ChatCompletionUserMessageParam(role="user", content=instruction)]
        )
        return await self._call_llm(ctx, task)

    async def _call_llm(self, session: Context, task: Task) -> str:
        while True:
            response = await self.llm.completion(
                messages=task.conversations(),
                tools=self._tool.tools,
                tool_choice="auto",
                max_tokens=4096,
            )
            await task.step([response.choices[0].message])

            response_message = response.choices[0].message
            tool_calls = response_message.get("tool_calls", None)

            if tool_calls is None:
                return response_message.content

            results = await self._tool.call(tool_calls, session)
            await task.step(results)


class BrowserAgentTool(AgentTool):
    _tool: BrowserTool

    def __init__(self, tool: BrowserTool, llm: LLM = None):
        super().__init__(tool, llm)

    async def get_screenshot(self) -> str | None:
        return await self._tool.get_screenshot()

    async def goto_blank(self):
        await self._tool.goto_blank()

    async def chat(
        self,
        ctx: Context,
        instruction: str,
    ) -> str:
        if not self._tool.use_screenshot:
            return await super().chat(ctx, instruction)

        screenshot = await self._tool.get_screenshot()

        if not screenshot:
            return await super().chat(ctx, instruction)

        task = Task(goal=instruction)
        ctx = ctx.fork(task)
        if self._tool.system_prompt:
            await task.step(
                [
                    ChatCompletionSystemMessageParam(
                        role="system", content=self._tool.system_prompt
                    )
                ]
            )

        await task.step(
            [
                ChatCompletionUserMessageParam(
                    role="user",
                    content=[
                        {
                            "type": "text",
                            "text": instruction,
                        },
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": screenshot,
                            },
                        },
                    ],
                )
            ]
        )

        return await self._call_llm(ctx, task)



import asyncio

from npiai import agent
from npiai.tools import GitHub


async def main():
    async with agent.wrap(GitHub()) as github:
        return await github.chat("Star and fork npi-ai/npi")


if __name__ == "__main__":
    asyncio.run(main())



import os

from twilio.rest import Client

from npiai import FunctionTool, function
from npiai.error.auth import UnauthorizedError


class Twilio(FunctionTool):
    name = "twilio"
    description = (
        'Send messages using Twilio, e.g., twilio("send a message to +1650705xxxx")'
    )
    system_prompt = "You are a Twilio Agent can send messages via WhatsApp or SMS"

    _sid: str | None
    _token: str | None
    _from_number: str | None
    _client: Client

    def __init__(
        self, account_sid: str = None, auth_token: str = None, from_number: str = None
    ):
        super().__init__()

        self._sid = account_sid or os.environ.get("TWILIO_ACCOUNT_SID", None)
        self._token = auth_token or os.environ.get("TWILIO_AUTH_TOKEN", None)
        self._from_number = from_number or os.environ.get("TWILIO_FROM_NUMBER", None)

    async def start(self):
        if not self._sid:
            raise UnauthorizedError("Twilio account sid is missing")
        if not self._token:
            raise UnauthorizedError("Twilio auth token is missing")

        self._client = Client(self._sid, self._token)

    @function
    async def send_sms_message(self, to: str, body: str) -> str:
        """
        Send a sms message to a phone number.

        Args:
            to: The phone number you want to send the message to.
            body: The body of the message.
        """
        message = self._client.messages.create(
            from_=f"{self._from_number}",
            body=body,
            to=to,
        )

        return f"Message has been successfully sent, sid: {message.sid}"

    @function
    async def send_whatsapp_message(self, to: str, body: str) -> str:
        """
        Send a whatsapp message to a phone number.

        Args:
            to: The phone number you want to send the message to.
            body: The body of the message.
        """
        message = self._client.messages.create(
            from_=f"whatsapp:{self._from_number}",
            body=body,
            to=f"whatsapp:{to}",
        )

        return f"Message has been successfully sent, sid: {message.sid}"



import functools
from typing import List, Callable

from langchain_core.tools import BaseToolkit, StructuredTool
from langchain_core.pydantic_v1 import create_model, PrivateAttr, Field

from npiai.core import BaseTool as NPiBaseTool
from npiai.context import Context


def unwrap_context(func: Callable, ctx_param_name: str | None) -> Callable:
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        if ctx_param_name:
            # create an empty context for tools
            kwargs[ctx_param_name] = Context()
        return await func(*args, **kwargs)

    return wrapper


class NPiLangChainToolkit(BaseToolkit):
    _tools: List[StructuredTool] = PrivateAttr()

    def __init__(self, npi_tool: NPiBaseTool):
        BaseToolkit.__init__(self)
        self._tools = []

        for fn_reg in npi_tool.unpack_functions():
            # remove ctx param
            fn = unwrap_context(fn_reg.fn, fn_reg.ctx_param_name)
            # recreate pydantic-v1 compatible model
            params = {}

            if fn_reg.model:
                for name, field in fn_reg.model.model_fields.items():
                    if field.default and not field.is_required():
                        field_v1 = Field(
                            default=field.default,
                            description=field.description,
                        )
                    else:
                        field_v1 = Field(description=field.description)
                    params[name] = (field.annotation, field_v1)

            schema_model_v1 = create_model(
                f"{fn_reg.name}__model",
                **params,
            )

            self._tools.append(
                StructuredTool.from_function(
                    coroutine=fn,
                    name=fn_reg.name,
                    description=fn_reg.description,
                    args_schema=schema_model_v1,
                    infer_schema=False,
                )
            )

    def get_tools(self) -> List[StructuredTool]:
        return self._tools



from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class FromContext:
    query: str
    constraints: str | None = None
    # generated attributes
    name: str = None
    return_type: Any = None



import asyncio
from npiai import agent
from npiai.tools.web.chromium import Chromium


async def main():
    async with agent.wrap(Chromium(headless=False)) as browser:
        return await browser.chat(
            "Book a one-way flight from ATL to LAX on 4/20 using Google Flights."
        )


if __name__ == "__main__":
    asyncio.run(main())



from .context import Context


class ContextManager:
    """the manager of the context"""

    def __init__(self):
        self.contexts = {}

    def get_context(self, ssid: str) -> Context | None:
        return self.contexts.get(ssid, None)

    def save_context(self, ctx: Context):
        self.contexts[ctx.id] = ctx

    def release(self, tid: str) -> None:
        """delete a context by id"""
        del self.contexts[tid]



from docstring_parser.google import GoogleParser, DEFAULT_SECTIONS, Section, SectionType

NPI_SECTIONS = [
    *DEFAULT_SECTIONS,
    Section("FewShots", "few_shots", SectionType.SINGULAR),
]


def parse_docstring(docstring: str):
    return GoogleParser(sections=NPI_SECTIONS).parse(docstring)



"""The basic interface for NPi Tools"""

import dataclasses
import inspect
import json
import re
from abc import ABC
from typing import Dict, List, Optional, Any, Type, Annotated, get_args, get_origin
from textwrap import dedent

import yaml

from pydantic import Field, create_model
from litellm.types.completion import (
    ChatCompletionToolMessageParam,
    ChatCompletionMessageParam,
)
from litellm.types.utils import ChatCompletionMessageToolCall
from openai.types.chat import ChatCompletionToolParam

from npiai.core.base import BaseTool, BaseFunctionTool
from npiai.core.hitl import HITL
from npiai.types import FunctionRegistration, ToolFunction, Shot, ToolMeta, FromContext
from npiai.utils import (
    logger,
    sanitize_schema,
    parse_docstring,
    to_async_fn,
    is_template_str,
)
from npiai.context import Context
from npiai.constant import CTX_QUERY_POSTFIX

__NPI_TOOL_ATTR__ = "__NPI_TOOL_ATTR__"


def function(
    tool_fn: ToolFunction = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    schema: Dict[str, Any] = None,
    model: Optional[Type[BaseTool]] = None,
    few_shots: Optional[List[Shot]] = None,
):
    """
    NPi Tool decorator for functions

    Args:
        tool_fn: Tool function. This value will be set automatically.
        name: Tool name. The tool function name will be used if not given.
        description: Tool description. This value will be inferred from the tool's docstring if not given.
        schema: Tool parameters schema. This value will be inferred from the tool's type hints if not given.
        model: Pydantic model for tool schema. This value will be built from the schema if not given.
        few_shots: Predefined working examples.

    Returns:
        Wrapped tool function that will be registered on the tools class
    """

    def decorator(fn: ToolFunction):
        setattr(
            fn,
            __NPI_TOOL_ATTR__,
            ToolMeta(
                name=name,
                description=description,
                schema=schema,
                model=model,
                few_shots=few_shots,
            ),
        )

        return fn

    # called as `@function`
    if callable(tool_fn):
        return decorator(tool_fn)

    # called as `@function(...)`
    return decorator


class FunctionTool(BaseFunctionTool, ABC):
    """The basic interface for the natural language programming interface"""

    _tools: List[ChatCompletionToolParam]
    _fn_map: Dict[str, FunctionRegistration]
    _sub_tools: List[BaseTool]

    _started: bool = False

    @property
    def functions(self):
        return list(self._fn_map.values())

    @property
    def sub_tools(self):
        return self._sub_tools

    def __init__(self):
        super().__init__()

        self._tools = []
        self._fn_map = {}
        self._sub_tools = []

        self._register_tools()

    def unpack_functions(self) -> List[FunctionRegistration]:
        return list(self._fn_map.values())

    def use_hitl(self, hitl: HITL):
        """
        Attach the given HITL handler to this tool and all its sub apps

        Args:
            hitl: HITL handler
        """
        super().use_hitl(hitl)

        for app in self._sub_tools:
            app.use_hitl(hitl)

    async def start(self):
        """Start the tools"""
        if not self._started:
            self._started = True
            for tool in self._sub_tools:
                await tool.start()

    async def end(self):
        """Stop and dispose the tools"""
        if self._started:
            self._started = False
            for app in self._sub_tools:
                await app.end()

    def add_tool(
        self,
        *tools: BaseTool,
    ):
        for tool in tools:
            # share hitl handler
            if tool._hitl is None:
                tool.use_hitl(self._hitl)

            self._sub_tools.append(tool)

            for fn_reg in tool.unpack_functions():
                scoped_fn_reg = dataclasses.replace(
                    fn_reg, name=f"{tool.name}_{fn_reg.name}"
                )
                self._add_function(scoped_fn_reg)

    async def debug(
        self,
        session: Context | None = None,
        app_name: str = None,
        fn_name: str = None,
        args: Dict[str, Any] = None,
    ) -> str:
        if app_name:
            fn_name = f"{app_name}_{fn_name}"

        return await self.exec(session, fn_name, args)

    async def call(
        self,
        tool_calls: List[ChatCompletionMessageToolCall],
        session: Context | None = None,
    ) -> List[ChatCompletionMessageParam]:
        if session is None:
            session = Context()
        results: List[ChatCompletionToolMessageParam] = []

        for call in tool_calls:
            fn_name = call.function.name
            args = json.loads(call.function.arguments)
            call_msg = f"[{self.name}]: Calling {fn_name}"

            if args:
                arg_list = ", ".join(f"{k}={json.dumps(v)}" for k, v in args.items())
                call_msg += f"({arg_list})"
            else:
                call_msg += "()"

            logger.info(call_msg)
            await session.send(call_msg)

            try:
                res = await self.exec(session, fn_name, args)
            except Exception as e:
                logger.error(e)
                res = f"Exception while executing {fn_name}: {e}"
                await session.send(res)

            logger.debug(f"[{self.name}]: function `{fn_name}` returned: {res}")

            results.append(
                ChatCompletionToolMessageParam(
                    role="tool",
                    tool_call_id=call.id,
                    content=str(res),
                )
            )

        return results

    def _register_tools(self):
        """
        Find the wrapped tool functions and register them in this tools
        """
        for attr, fn in inspect.getmembers(
            self, lambda x: inspect.ismethod(x) or inspect.isfunction(x)
        ):
            tool_meta: ToolMeta | None = getattr(fn, __NPI_TOOL_ATTR__, None)

            if not callable(fn) or not tool_meta:
                continue

            params = list(inspect.signature(fn).parameters.values())
            docstr = parse_docstring(inspect.getdoc(fn))
            tool_name = tool_meta.name or fn.__name__
            tool_desc = tool_meta.description or docstr.description

            if not tool_desc:
                raise ValueError(
                    f"Unable to get the description of tool function `{fn}`"
                )

            # parse schema
            tool_model = tool_meta.model
            tool_schema = tool_meta.schema
            ctx_param_name = None
            ctx_variables = []

            if not tool_model and len(params) > 0:
                # get parameter descriptions
                param_descriptions = {}
                for p in docstr.params:
                    param_descriptions[p.arg_name] = p.description

                # get parameter field definitions
                param_fields = {}
                for p in params:
                    if p.annotation is Context:
                        ctx_param_name = p.name
                        continue

                    if get_origin(p.annotation) is Annotated:
                        # extract context variables
                        return_type, anno = get_args(p.annotation)
                        if isinstance(anno, FromContext):
                            ctx_variables.append(
                                dataclasses.replace(
                                    anno,
                                    name=p.name,
                                    return_type=return_type,
                                )
                            )

                            if is_template_str(anno.query):
                                param_fields[f"{p.name}{CTX_QUERY_POSTFIX}"] = (
                                    str,
                                    Field(
                                        default=anno.query,
                                        description=dedent(
                                            f"""
                                            This parameter is a query to retrieve information from the memory storage.
                                            For the following query, you should replace the strings surrounded by braces
                                            `{{}}` with the information from current context.
                                            
                                            Query: {anno.query}
                                            """
                                        ),
                                    ),
                                )
                            continue

                    param_fields[p.name] = (
                        p.annotation,
                        Field(
                            default=(
                                p.default
                                if p.default is not inspect.Parameter.empty
                                else ...
                            ),
                            description=param_descriptions.get(p.name, ""),
                        ),
                    )

                if len(param_fields):
                    tool_model = create_model(f"{tool_name}_model", **param_fields)

            if not tool_schema and tool_model:
                tool_schema = sanitize_schema(tool_model)

            # parse examples
            tool_few_shots = tool_meta.few_shots
            doc_shots = [m for m in docstr.meta if m.args == ["few_shots"]]

            if not tool_few_shots and len(doc_shots) > 0:
                tool_few_shots = []

                for shot in doc_shots:
                    items = re.findall(r"^\s*- ", shot.description, re.MULTILINE)
                    if len(items) == 1:
                        # remove leading '- ' to avoid indentation issues
                        shot_data = yaml.safe_load(
                            re.sub(r"^\s*- ", "", shot.description)
                        )
                    else:
                        shot_data = yaml.safe_load(shot.description)

                    if not isinstance(shot_data, list):
                        shot_data = [shot_data]

                    for d in shot_data:
                        tool_few_shots.append(Shot(**d))

            self._add_function(
                FunctionRegistration(
                    # wrap fn in an async wrapper
                    fn=to_async_fn(fn),
                    name=tool_name,
                    ctx_variables=ctx_variables,
                    ctx_param_name=ctx_param_name,
                    description=tool_desc.strip(),
                    schema=tool_schema,
                    model=tool_model,
                    few_shots=tool_few_shots,
                )
            )

    def _add_function(self, fn_reg: FunctionRegistration):
        if fn_reg.name in self._fn_map:
            raise Exception(f"Duplicate function: {fn_reg.name}")

        self._fn_map[fn_reg.name] = fn_reg
        self._tools.append(fn_reg.get_tool_param())
